#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2023, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for lcd document filter'''

# pylint: disable=R0201,C0115,C0116

import unittest
from ttconv.filters.doc.lcd import LCDDocFilter, LCDDocFilterConfig
import ttconv.model as model
import ttconv.style_properties as styles

class LCDFilterTests(unittest.TestCase):

  '''
  <region xml:id="r1"/>
  <region xml:id="r2"/>
  <region xml:id="r2" begin="2s" end="9s">
    <set tts:color="red"/>
  </region>

  <body>
    <div>

    </div>
  </body>
  '''

  def test_region_merging(self):
    doc = model.ContentDocument()

    # r1
    r1 = model.Region("r1", doc)
    doc.put_region(r1)

    # r2: tts:extent="100% 100%"
    r2 = model.Region("r2", doc)
    r2.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        height=styles.LengthType(value=100),
        width=styles.LengthType(value=100)
      )
    )
    doc.put_region(r2)

    # r3: tts:displayAlign="after"
    r3 = model.Region("r3", doc)
    r3.set_style(
      styles.StyleProperties.DisplayAlign,
      styles.DisplayAlignType.after
    )
    doc.put_region(r3)

    # r4: set: tts:displayAlign="after"
    r4 = model.Region("r4", doc)
    r4.add_animation_step(
      model.DiscreteAnimationStep(
        style_property=styles.StyleProperties.DisplayAlign,
        begin=None,
        end=None,
        value=styles.DisplayAlignType.after
      )
    )
    doc.put_region(r4)

    # r5: begin="1s"
    r5 = model.Region("r5", doc)
    r5.set_begin(1)
    doc.put_region(r5)

    # body
    body = model.Body(doc)
    doc.set_body(body)

    # div
    div = model.Div(doc)
    body.push_child(div)

    # p1: r1
    p1 = model.P(doc)
    p1.set_id("p1")
    p1.set_region(r1)
    div.push_child(p1)

    # p2: r2
    p2 = model.P(doc)
    p2.set_id("p2")
    p2.set_region(r2)
    div.push_child(p2)

    # p3: r3
    p3 = model.P(doc)
    p3.set_id("p3")
    p3.set_region(r3)
    div.push_child(p3)

    # p4: r4
    p4 = model.P(doc)
    p4.set_id("p4")
    p4.set_region(r4)
    div.push_child(p4)

    # p5: r5
    p5 = model.P(doc)
    p5.set_id("p5")
    p5.set_region(r5)
    div.push_child(p5)

    filter = LCDDocFilter(LCDDocFilterConfig())

    self.assertIsNone(filter.process(doc))

    self.assertSetEqual(
      set(["r1", "r3", "r5"]),
      set([r.get_id() for r in doc.iter_regions()])
    )

  def test_region_resizing(self):
    doc = model.ContentDocument()

    # r1: origin=10,10 extent=80,20
    r1 = model.Region("r1", doc)
    r1.set_style(
      styles.StyleProperties.Origin,
      styles.CoordinateType(
        x=styles.LengthType(value=10),
        y=styles.LengthType(value=10)
      )
    )
    r1.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        height=styles.LengthType(value=20),
        width=styles.LengthType(value=80)
      )
    )
    doc.put_region(r1)

    # r2: origin=10,70 extent=80,20
    r2 = model.Region("r2", doc)
    r2.set_style(
      styles.StyleProperties.Origin,
      styles.CoordinateType(
        x=styles.LengthType(value=10),
        y=styles.LengthType(value=70)
      )
    )
    r2.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        height=styles.LengthType(value=20),
        width=styles.LengthType(value=80)
      )
    )
    doc.put_region(r2)

    # r3: origin=10,10 extent=80,20 displayAlign=after
    r3 = model.Region("r3", doc)
    r3.set_style(
      styles.StyleProperties.Origin,
      styles.CoordinateType(
        x=styles.LengthType(value=10),
        y=styles.LengthType(value=10)
      )
    )
    r3.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        height=styles.LengthType(value=20),
        width=styles.LengthType(value=80)
      )
    )
    r3.set_style(
      styles.StyleProperties.DisplayAlign,
      styles.DisplayAlignType.after
    )
    doc.put_region(r3)

    # r4: origin=10,70 extent=80,20 displayAlign=after
    r4 = model.Region("r4", doc)
    r4.set_style(
      styles.StyleProperties.Origin,
      styles.CoordinateType(
        x=styles.LengthType(value=10),
        y=styles.LengthType(value=70)
      )
    )
    r4.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        height=styles.LengthType(value=20),
        width=styles.LengthType(value=80)
      )
    )
    r4.set_style(
      styles.StyleProperties.DisplayAlign,
      styles.DisplayAlignType.after
    )
    doc.put_region(r4)

    # body
    body = model.Body(doc)
    doc.set_body(body)

    # div
    div = model.Div(doc)
    body.push_child(div)

    # p1: r1
    p1 = model.P(doc)
    p1.set_id("p1")
    p1.set_region(r1)
    div.push_child(p1)

    # p2: r2
    p2 = model.P(doc)
    p2.set_id("p2")
    p2.set_region(r2)
    div.push_child(p2)

    # p3: r3
    p3 = model.P(doc)
    p3.set_id("p3")
    p3.set_region(r3)
    div.push_child(p3)

    # p4: r4
    p4 = model.P(doc)
    p4.set_id("p4")
    p4.set_region(r4)
    div.push_child(p4)

    # apply filter
    filter = LCDDocFilter(LCDDocFilterConfig())

    self.assertIsNone(filter.process(doc))

    self.assertSetEqual(
      set(["r1", "r2"]),
      set([r.get_id() for r in doc.iter_regions()])
    )

    self.assertLessEqual(
      doc.get_region("r1").get_style(styles.StyleProperties.Origin).y.value,
      50
    )

    self.assertGreaterEqual(
      doc.get_region("r1").get_style(styles.StyleProperties.Extent).height.value,
      50
    )

    self.assertEqual(
      doc.get_region("r1").get_style(styles.StyleProperties.DisplayAlign),
      styles.DisplayAlignType.before
    )

    self.assertLessEqual(
      doc.get_region("r2").get_style(styles.StyleProperties.Origin).y.value,
      50
    )

    self.assertGreaterEqual(
      doc.get_region("r2").get_style(styles.StyleProperties.Extent).height.value,
      50
    )

    self.assertEqual(
      doc.get_region("r2").get_style(styles.StyleProperties.DisplayAlign),
      styles.DisplayAlignType.after
    )

  def test_text_align(self):
    doc = model.ContentDocument()

    # r1: origin=10,10 extent=80,20
    r1 = model.Region("r1", doc)
    doc.put_region(r1)

    # body
    body = model.Body(doc)
    doc.set_body(body)

    # div
    div = model.Div(doc)
    body.push_child(div)

    # p1: r1
    p1 = model.P(doc)
    p1.set_id("p1")
    p1.set_region(r1)
    p1.set_style(styles.StyleProperties.TextAlign, styles.TextAlignType.end)
    div.push_child(p1)

    # apply filter
    filter = LCDDocFilter(LCDDocFilterConfig())

    self.assertIsNone(filter.process(doc))

    self.assertIsNone(p1.get_style(styles.StyleProperties.TextAlign))
    self.assertEqual(body.get_style(styles.StyleProperties.TextAlign), styles.TextAlignType.center)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Time Code"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from fractions import Fraction
from math import floor, ceil
from typing import Optional, Union


class _HHMMSSTimeExpression(ABC):
  """Time code basic definition"""

  def __init__(self, hours: int, minutes: int, seconds: int):
    self._hours = hours
    self._minutes = minutes
    self._seconds = seconds

  def get_hours(self) -> int:
    """Returns time code hours"""
    return self._hours

  def get_minutes(self) -> int:
    """Returns time code minutes"""
    return self._minutes

  def get_seconds(self) -> int:
    """Returns time code seconds"""
    return self._seconds

class ClockTime(_HHMMSSTimeExpression):
  """Millisecond-based time code definition"""

  TIME_CODE_PATTERN = ':'.join(['(?P<h>[0-9]{2,})',
                                '(?P<m>[0-9]{2})',
                                '(?P<s>[0-9]{2})']) + \
                      '(.|,)' + '(?P<ms>[0-9]{2,3})'

  def __init__(self, hours: int, minutes: int, seconds: int, milliseconds: int):
    super().__init__(hours, minutes, seconds)
    self._milliseconds = milliseconds
    self._ms_separator = "."

  def get_milliseconds(self) -> int:
    """Returns time code milliseconds"""
    return self._milliseconds

  def to_seconds(self) -> Fraction:
    """Converts current time code to seconds"""
    return self.get_hours() * 3600 + \
           self.get_minutes() * 60 + \
           self.get_seconds() + \
           Fraction(self._milliseconds, 1000)

  @staticmethod
  def parse(time_code: str) -> ClockTime:
    """Reads the time code string and converts to a ClockTime instance"""
    tc_regex = re.compile(ClockTime.TIME_CODE_PATTERN)
    match = tc_regex.match(time_code)

    if match is not None:
      return ClockTime(int(match.group('h')),
                       int(match.group('m')),
                       int(match.group('s')),
                       int(match.group('ms')))

    raise ValueError("Invalid time code format")

  @staticmethod
  def from_seconds(seconds: Union[float, Fraction]) -> ClockTime:
    """Creates a time code from time offset in seconds (rounds to the closest millisecond)"""

    if seconds < 0:
      raise ValueError("Seconds must not be less than zero")

    seconds = round(seconds, 3)

    h = floor(seconds / 3600)
    m = floor(seconds / 60 % 60)
    s = floor(seconds % 60)
    ms = round((seconds % 1) * 1000)

    return ClockTime(h, m, s, ms)

  def set_separator(self, separator: str):
    """Sets separator character before the millisecond digits into the literal representation of this time code"""
    self._ms_separator = separator

  def __str__(self):
    return ":".join(f'{item:02d}' for item in [self._hours, self._minutes, self._seconds]) \
           + self._ms_separator + f'{self._milliseconds:03}'

  def __eq__(self, other: ClockTime):
    if not isinstance(other, ClockTime):
      return False
    return self._hours == other.get_hours() and \
           self._minutes == other.get_minutes() and \
           self._seconds == other.get_seconds() and \
           self._milliseconds == other.get_milliseconds()


FPS_23_98 = Fraction(24000, 1001)
FPS_24 = Fraction(24, 1)
FPS_25 = Fraction(25, 1)
FPS_29_97 = Fraction(30000, 1001)
FPS_30 = Fraction(30, 1)
FPS_50 = Fraction(50, 1)
FPS_59_94 = Fraction(60000, 1001)
FPS_60 = Fraction(60, 1)


class SmpteTimeCode(_HHMMSSTimeExpression):
  """Frame-based time code definition"""

  _NDF_TC_PATTERN = ':'.join(['(?P<ndf_h>[0-9]{2})',
                                          '(?P<ndf_m>[0-9]{2})',
                                          '(?P<ndf_s>[0-9]{2})',
                                          '(?P<ndf_f>[0-9]{2})'])
  
  _NDF_TC_RE = re.compile(_NDF_TC_PATTERN)

  _DF_TC_PATTERN = '(:|;|.|,)'.join(['(?P<df_h>[0-9]{2})',
                                                 '(?P<df_m>[0-9]{2})',
                                                 '(?P<df_s>[0-9]{2})',
                                                 '(?P<df_f>[0-9]{2})'])
  
  DF_TC_RE = re.compile(_DF_TC_PATTERN)

  def __init__(self, hours: int, minutes: int, seconds: int, frames: int, frame_rate: Fraction, is_df: Optional[bool] = None):
    super().__init__(hours, minutes, seconds)
    self._frames: int = frames
    self._frame_rate: Fraction = frame_rate

    if is_df is None:
       self._is_drop_frame = frame_rate.denominator == 1001
    else:
      if is_df and frame_rate.denominator != 1001:
        raise ValueError("The numerator of the frame rate of a drop frame timecode must be 1001")
      self._is_drop_frame: bool = is_df

  def get_frames(self) -> int:
    """Returns time code frames"""
    return self._frames

  def get_frame_rate(self) -> Fraction:
    """Returns time code frame rate"""
    return self._frame_rate

  def to_frames(self) -> int:
    """Converts current time code into a number of frames"""
    dropped_frames = 0

    fps = ceil(self._frame_rate)

    if self.is_drop_frame():
      drop_frames_per_minute = round(60 * (fps - self._frame_rate))  # 2 at 29.97 fps

      nb_of_minute_tens = self._hours * 6 + floor(self._minutes / 10)
      nb_of_drop_frames_in_tens = drop_frames_per_minute * 9 * nb_of_minute_tens
      remaining_minutes = self._minutes % 10
      nb_of_drop_frames_in_remaining = drop_frames_per_minute * remaining_minutes

      dropped_frames = nb_of_drop_frames_in_tens + nb_of_drop_frames_in_remaining

    whole_seconds = self.get_hours() * 3600 + \
        self.get_minutes() * 60 + \
        self.get_seconds()
    
    return int(whole_seconds * fps) + self._frames - dropped_frames

  def to_seconds(self) -> Fraction:
    """Converts current time code into seconds"""
    return self.to_frames() / self._frame_rate

  def to_temporal_offset(self) -> Fraction:
    """Converts current time code into a second-based fraction"""
    nb_frames = self.to_frames()
    return Fraction(nb_frames, self._frame_rate)

  def is_drop_frame(self) -> bool:
    """Returns whether the time code is drop-frame or not"""
    return self._is_drop_frame

  def add_frames(self, nb_frames=1):
    """Add frames to the current time code"""
    frames = self.to_frames() + nb_frames

    new_time_code = SmpteTimeCode.from_frames(frames, self._frame_rate)

    self._hours = new_time_code.get_hours()
    self._minutes = new_time_code.get_minutes()
    self._seconds = new_time_code.get_seconds()
    self._frames = new_time_code.get_frames()

  @staticmethod
  def parse(time_code: str, frame_rate: Fraction) -> SmpteTimeCode:
    """Reads the time code string and converts to a SmpteTimeCode instance"""
    
    match = SmpteTimeCode._NDF_TC_RE.match(time_code)

    if match is not None:
      # NDF timecode
      return SmpteTimeCode(int(match.group('ndf_h')),
                           int(match.group('ndf_m')),
                           int(match.group('ndf_s')),
                           int(match.group('ndf_f')),
                           frame_rate,
                           False)

    if frame_rate.denominator != 1001:
      frame_rate = frame_rate * Fraction(1000, 1001)

    match = SmpteTimeCode.DF_TC_RE.match(time_code)

    if match is not None:
      return SmpteTimeCode(int(match.group('df_h')),
                           int(match.group('df_m')),
                           int(match.group('df_s')),
                           int(match.group('df_f')),
                           frame_rate,
                           True)

    raise ValueError("Invalid time code format")

  @staticmethod
  def from_frames(nb_frames: Union[int, Fraction], frame_rate: Fraction, is_df: Optional[bool] = None) -> SmpteTimeCode:
    """Creates a time code from a number of frames and a frame rate"""
    if frame_rate is None:
      raise ValueError("Cannot compute time code from frames without frame rate")

    if nb_frames is None or nb_frames < 0:
      raise ValueError("Number of frames must not None and must not be less than zero")

    if is_df is None:
       drop_frame: bool = frame_rate.denominator == 1001
    else:
      if is_df and frame_rate.denominator != 1001:
        raise ValueError("The numerator of the frame rate of a drop frame timecode must be 1001")
      drop_frame: bool = is_df

    fps = ceil(frame_rate)

    if drop_frame:
      # add two dropped frames every minute, but not when the minute count is divisible by 10

      nb_frames_in_one_minute = 60 * frame_rate  # 1798 at 29.97 fps
      nb_frames_in_ten_minutes = round(10 * nb_frames_in_one_minute)  # 17982 at 29.97 fps
      drop_frames_per_minute = round(60 * (fps - frame_rate))  # 2 at 29.97 fps

      nb_of_minute_tens = floor(nb_frames / nb_frames_in_ten_minutes)
      nb_of_remaining_frames = round(nb_frames % nb_frames_in_ten_minutes)

      nb_of_drop_frames_in_tens = drop_frames_per_minute * 9 * nb_of_minute_tens

      nb_of_remaining_minutes = floor((nb_of_remaining_frames - drop_frames_per_minute) / round(nb_frames_in_one_minute))

      if nb_of_remaining_minutes < 0:
        nb_of_remaining_minutes = 0

      nb_of_drop_frames_in_minutes = nb_of_remaining_minutes * drop_frames_per_minute

      nb_frames += nb_of_drop_frames_in_tens + nb_of_drop_frames_in_minutes

    h = floor(nb_frames / (60 * 60 * fps))
    m = floor(nb_frames / (60 * fps)) % 60
    s = floor(nb_frames / fps) % 60
    f = ceil(nb_frames % fps)

    return SmpteTimeCode(h, m, s, f, frame_rate, drop_frame)

  @staticmethod
  def from_seconds(seconds: Union[float, Fraction], frame_rate: Fraction) -> SmpteTimeCode:
    """Creates a time code from time offset in seconds and a frame rate"""
    if frame_rate is None:
      raise ValueError("Cannot compute SMPTE time code from seconds without frame rate")

    return SmpteTimeCode.from_frames(int(seconds * frame_rate), frame_rate)

  def __str__(self):
    if self.is_drop_frame():
      return ":".join(f'{item:02}' for item in [self._hours, self._minutes, self._seconds]) + ";" + f'{self._frames:02}'
    return ":".join(f'{item:02}' for item in [self._hours, self._minutes, self._seconds, self._frames])

  def __repr__(self):
    return f"{self} at {float(self._frame_rate)} fps"

  def __eq__(self, other: SmpteTimeCode):
    if not isinstance(other, SmpteTimeCode):
      return False
    return self._hours == other.get_hours() and \
           self._minutes == other.get_minutes() and \
           self._seconds == other.get_seconds() and \
           self._frames == other.get_frames() and \
           self._is_drop_frame == other.is_drop_frame()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Process TTML elements'''

from __future__ import annotations
import logging
from fractions import Fraction
import typing
import numbers
import xml.etree.ElementTree as et
from ttconv.imsc.config import ContentProfilesSignaling
import ttconv.model as model
import ttconv.style_properties as model_styles
import ttconv.imsc.namespaces as xml_ns
import ttconv.imsc.attributes as imsc_attr
from ttconv.imsc.style_properties import StyleProperties
import ttconv.imsc.style_properties as imsc_styles


LOGGER = logging.getLogger(__name__)

class TTMLElement:
  '''Static information about a TTML element
  '''

  class ParsingContext(imsc_styles.StyleParsingContext):
    '''State information when parsing a TTML element'''

    def __init__(self, ttml_class: typing.Type[TTMLElement], parent_ctx: typing.Optional[TTMLElement.ParsingContext] = None):

      self.doc = parent_ctx.doc if parent_ctx is not None else model.ContentDocument()

      self.style_elements: typing.Dict[str, StyleElement] = parent_ctx.style_elements if parent_ctx is not None else {}

      self.temporal_context = parent_ctx.temporal_context if parent_ctx is not None else imsc_attr.TemporalAttributeParsingContext()

      self.ttml_class: typing.Type[TTMLElement] = ttml_class

      self.lang: typing.Optional[str] = None

      self.space: typing.Optional[model.WhiteSpaceHandling] = None

      self.time_container: imsc_attr.TimeContainer = imsc_attr.TimeContainer.par

      self.explicit_begin: typing.Optional[Fraction] = None

      self.implicit_begin: typing.Optional[Fraction] = None

      self.desired_begin: typing.Optional[Fraction] = None

      self.explicit_end: typing.Optional[Fraction] = None

      self.implicit_end: typing.Optional[Fraction] = None

      self.desired_end: typing.Optional[Fraction] = None

      self.explicit_dur: typing.Optional[Fraction] = None

    def process_lang_attribute(self, parent_ctx: TTMLElement.ParsingContext, xml_elem):
      '''Processes the xml:lang attribute, including inheritance from the parent
      '''
      lang_attr_value = imsc_attr.XMLLangAttribute.extract(xml_elem)
      self.lang = lang_attr_value if lang_attr_value is not None else parent_ctx.lang

    def process_space_attribute(self, parent_ctx: TTMLElement.ParsingContext, xml_elem):
      '''Processes the xml:space attribute, including inheritance from the parent
      '''
      space_attr_value = imsc_attr.XMLSpaceAttribute.extract(xml_elem)
      self.space = space_attr_value if space_attr_value is not None else parent_ctx.space

  class WritingContext:
    '''State information when writing a TTML element'''

    def __init__(self, frame_rate: Fraction, time_expression_syntax: imsc_attr.TimeExpressionSyntaxEnum):
      self.temporal_context = imsc_attr.TemporalAttributeWritingContext(
        frame_rate=frame_rate,
        time_expression_syntax=time_expression_syntax
        )

  @staticmethod
  def is_instance(xml_elem) -> bool:
    '''Returns true if the XML element `xml_elem` is an instance of the class
    '''
    raise NotImplementedError


class TTElement(TTMLElement):
  '''Processes the TTML <tt> element
  '''

  class ParsingContext(TTMLElement.ParsingContext):
    '''State information when parsing a <tt> element'''

  qn = f"{{{xml_ns.TTML}}}tt"

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == TTElement.qn

  @staticmethod
  def from_xml(
    _parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element,
    progress_callback: typing.Callable[[numbers.Real], typing.NoReturn] = None
  ) -> TTElement.ParsingContext:
    '''`_parent_ctx` is ignored and can be set to `None`
    '''

    tt_ctx = TTElement.ParsingContext(TTElement)

    # process attributes

    space_attr = imsc_attr.XMLSpaceAttribute.extract(xml_elem)

    tt_ctx.space = space_attr if space_attr is not None else model.WhiteSpaceHandling.DEFAULT

    lang_attr = imsc_attr.XMLLangAttribute.extract(xml_elem)

    if lang_attr is None:
      LOGGER.warning("xml:lang not specified on tt")
      lang_attr = ""

    tt_ctx.lang = lang_attr

    tt_ctx.doc.set_lang(tt_ctx.lang)

    tt_ctx.doc.set_cell_resolution(
      imsc_attr.CellResolutionAttribute.extract(xml_elem)
    )

    tt_ctx.doc.set_content_profiles(
      imsc_attr.ContentProfilesAttribute.extract(xml_elem)
    )

    tt_ctx.temporal_context.time_base = imsc_attr.TimeBaseAttribute.extract(xml_elem)
    tt_ctx.temporal_context.drop_mode = imsc_attr.DropModeAttribute.extract(xml_elem)

    px_resolution = imsc_attr.ExtentAttribute.extract(xml_elem)

    if px_resolution is not None:
      tt_ctx.doc.set_px_resolution(px_resolution)

    active_area = imsc_attr.ActiveAreaAttribute.extract(xml_elem)

    if active_area is not None:
      tt_ctx.doc.set_active_area(active_area)

    ittp_aspect_ratio = imsc_attr.AspectRatioAttribute.extract(xml_elem)

    ttp_dar = imsc_attr.DisplayAspectRatioAttribute.extract(xml_elem)

    if ttp_dar is not None:

      tt_ctx.doc.set_display_aspect_ratio(ttp_dar)

    elif ittp_aspect_ratio is not None:

      tt_ctx.doc.set_display_aspect_ratio(ittp_aspect_ratio)

    if ittp_aspect_ratio is not None and ttp_dar is not None:

      LOGGER.warning("Both ittp:aspectRatio and ttp:displayAspectRatio specified on tt")

    tt_ctx.temporal_context.frame_rate = imsc_attr.FrameRateAttribute.extract(xml_elem)

    tt_ctx.temporal_context.tick_rate = imsc_attr.TickRateAttribute.extract(xml_elem)

    # process head and body children elements

    has_body = False
    has_head = False

    for child_element in xml_elem:

      if BodyElement.is_instance(child_element):

        if not has_body:

          has_body = True

          body_element = ContentElement.from_xml(tt_ctx, child_element)

          tt_ctx.doc.set_body(body_element.model_element if body_element is not None else None)

          progress_callback(1)

        else:
          LOGGER.error("More than one body element present")

      elif HeadElement.is_instance(child_element):

        if not has_head:

          has_head = True

          HeadElement.from_xml(tt_ctx, child_element)

          progress_callback(0.5)

        else:
          LOGGER.error("More than one head element present")

    return tt_ctx

  @staticmethod
  def from_model(
    model_doc: model.ContentDocument,
    frame_rate: typing.Optional[Fraction],
    time_expression_syntax: imsc_attr.TimeExpressionSyntaxEnum,
    progress_callback: typing.Callable[[numbers.Real], typing.NoReturn],
    content_profiles_signaling: ContentProfilesSignaling
  ) -> et.Element:
    '''Converts the data model to an IMSC document contained in an ElementTree Element'''

    ctx = TTMLElement.WritingContext(frame_rate, time_expression_syntax)

    tt_element = et.Element(TTElement.qn)

    imsc_attr.XMLLangAttribute.set(tt_element, model_doc.get_lang())

    if model_doc.get_cell_resolution() != model.CellResolutionType(rows=15, columns=32):
      imsc_attr.CellResolutionAttribute.set(tt_element, model_doc.get_cell_resolution())

    if content_profiles_signaling == ContentProfilesSignaling.CONTENT_PROFILES and \
      model_doc.get_content_profiles() is not None:
      imsc_attr.ContentProfilesAttribute.set(tt_element, model_doc.get_content_profiles())

    has_px = False

    all_elements = list(model_doc.iter_regions())

    if model_doc.get_body() is not None:
      all_elements.extend(model_doc.get_body().dfs_iterator())

    for element in all_elements:
      for model_style_prop in element.iter_styles():
        if StyleProperties.BY_MODEL_PROP[model_style_prop].has_px(element.get_style(model_style_prop)):
          has_px = True
          break
      for animation_step in element.iter_animation_steps():
        if StyleProperties.BY_MODEL_PROP[animation_step.style_property].has_px(animation_step.value):
          has_px = True
          break      
      if has_px:
        break

    if model_doc.get_px_resolution() is not None and has_px:
      imsc_attr.ExtentAttribute.set(tt_element, model_doc.get_px_resolution())

    if model_doc.get_active_area() is not None:
      imsc_attr.ActiveAreaAttribute.set(tt_element, model_doc.get_active_area())

    if model_doc.get_display_aspect_ratio() is not None:
      imsc_attr.DisplayAspectRatioAttribute.set(tt_element, model_doc.get_display_aspect_ratio())

    if frame_rate is not None:
      imsc_attr.FrameRateAttribute.set(tt_element, frame_rate)

    # Write the <head> section first
    head_element = HeadElement.from_model(ctx, model_doc)

    progress_callback(0.5)

    if head_element is not None:
      tt_element.append(head_element)

    model_body = model_doc.get_body()

    if model_body is not None:

      body_element = BodyElement.from_model(ctx, model_body)

      if body_element is not None:
        tt_element.append(body_element)

    progress_callback(1.0)

    return tt_element

class HeadElement(TTMLElement):
  '''Processes the TTML <head> element
  '''

  class ParsingContext(TTMLElement.ParsingContext):
    '''Maintains state when parsing a <head> element
    '''

  qn = f"{{{xml_ns.TTML}}}head"

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == HeadElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> HeadElement.ParsingContext:
    '''Converts the XML element `xml_elem` into its representation in the data model.
    `parent_ctx` contains state information passed from parent to child in the TTML hierarchy.
    '''
    
    head_ctx = HeadElement.ParsingContext(HeadElement, parent_ctx)

    # process attributes

    head_ctx.process_lang_attribute(parent_ctx, xml_elem)

    head_ctx.process_space_attribute(parent_ctx, xml_elem)

    # process layout and styling children elements

    has_layout = False
    has_styling = False
    
    for child_element in xml_elem:

      if LayoutElement.is_instance(child_element):

        if not has_layout:

          has_layout = True

          LayoutElement.from_xml(
            head_ctx,
            child_element
          )

        else:

          LOGGER.error("Multiple layout elements")

      elif StylingElement.is_instance(child_element):

        if not has_styling:

          has_styling = True

          StylingElement.from_xml(
            head_ctx,
            child_element
          )

        else:

          LOGGER.error("Multiple styling elements")

    return head_ctx

  @staticmethod
  def from_model(
    ctx: TTMLElement.WritingContext,
    model_doc: model.ContentDocument,
  )-> typing.Optional[et.Element]:
    '''Converts the ContentDocument `model_doc` into its TTML representation, i.e. an XML element.
    `ctx` contains state information used in the process.
    '''

    head_element = None

    styling_element = StylingElement.from_model(ctx, model_doc)

    if styling_element is not None:
      if head_element is None:
        head_element = et.Element(HeadElement.qn)
      head_element.append(styling_element)

    layout_element = LayoutElement.from_model(ctx, model_doc)

    if layout_element is not None:
      if head_element is None:
        head_element = et.Element(HeadElement.qn)
      head_element.append(layout_element)

    return head_element


class LayoutElement(TTMLElement):
  '''Process the TTML <layout> element
  '''
  
  class ParsingContext(TTMLElement.ParsingContext):
    '''Maintains state when parsing a <layout> element
    '''

  qn = f"{{{xml_ns.TTML}}}layout"

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == LayoutElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[LayoutElement.ParsingContext]:
    '''Converts the XML element `xml_elem` into its representation in the data model.
    `parent_ctx` contains state information passed from parent to child in the TTML hierarchy.
    '''

    layout_ctx = LayoutElement.ParsingContext(LayoutElement, parent_ctx)

    # process attributes

    layout_ctx.process_lang_attribute(parent_ctx, xml_elem)

    layout_ctx.process_space_attribute(parent_ctx, xml_elem)

    # process region elements
    
    for child_element in xml_elem:

      if RegionElement.is_instance(child_element):

        r = RegionElement.from_xml(layout_ctx, child_element)

        if r is not None:
          layout_ctx.doc.put_region(r.model_element)

      else:

        LOGGER.warning("Unexpected child of layout element")

    return layout_ctx

  @staticmethod
  def from_model(
    ctx: TTMLElement.WritingContext,
    model_doc: model.ContentDocument,
  ) -> typing.Optional[et.Element]:
    '''Returns a TTML `layout` element (an XML element) using the information in the ContentDocument `model_doc`.
    `ctx` contains state information used in the process.
    '''

    layout_element = None
    
    for r in model_doc.iter_regions():
      region_element = RegionElement.from_model(ctx, r)
      if region_element is not None:
        if layout_element is None:
          layout_element = et.Element(LayoutElement.qn)
        layout_element.append(region_element)

    return layout_element

class StylingElement(TTMLElement):
  '''Process the TTML <styling> element
  '''

  class ParsingContext(TTMLElement.ParsingContext):
    '''Maintains state when parsing a <styling> element
    '''

    def merge_chained_styles(self, style_element: StyleElement):
      '''Flattens Chained Referential Styling of the target `style_element` by specifying
      the style properties of the referenced style elements directly in the target element
      '''
      while len(style_element.style_refs) > 0:

        style_ref = style_element.style_refs.pop()

        if style_ref not in self.style_elements:
          LOGGER.error("Style id not present")
          continue

        self.merge_chained_styles(self.style_elements[style_ref])

        for style_prop, value in self.style_elements[style_ref].styles.items():
          style_element.styles.setdefault(style_prop, value)


  qn = f"{{{xml_ns.TTML}}}styling"

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == StylingElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[StylingElement.ParsingContext]:
    '''Converts the XML element `xml_elem` into its representation in the data model.
    `parent_ctx` contains state information passed from parent to child in the TTML hierarchy.
    '''

    styling_ctx = StylingElement.ParsingContext(StylingElement, parent_ctx)

    # process style and initial children elements

    for child_xml_elem in xml_elem:

      if InitialElement.is_instance(child_xml_elem):

        InitialElement.from_xml(styling_ctx, child_xml_elem)

      elif StyleElement.is_instance(child_xml_elem):
        
        style_element = StyleElement.from_xml(styling_ctx, child_xml_elem)

        if style_element is None:
          continue

        if style_element.id in styling_ctx.style_elements:
          LOGGER.error("Duplicate style id")
          continue

        style_element.style_elements[style_element.id] = style_element
      
    # merge style elements (the data model does not support referential
    # styling)

    for style_element in parent_ctx.style_elements.values():
      styling_ctx.merge_chained_styles(style_element)

    return styling_ctx

  @staticmethod
  def from_model(
    _ctx: TTMLElement.WritingContext,
    model_doc: model.ContentDocument
  ) -> typing.Optional[et.Element]:
    '''Returns a TTML `styling` element using the information in the ContentDocument `model_doc`.
    `ctx` contains state information used in the process.
    '''

    styling_element = None

    for style_prop, style_value in model_doc.iter_initial_values():

      imsc_style_prop = imsc_styles.StyleProperties.BY_MODEL_PROP.get(style_prop)

      if imsc_style_prop is None:
        LOGGER.error("Unknown property")
        continue

      initial_element = InitialElement.from_model(imsc_style_prop, style_value)
      if initial_element is not None:
        if styling_element is None:
          styling_element = et.Element(StylingElement.qn)
        styling_element.append(initial_element)

    return styling_element


class StyleElement(TTMLElement):
  '''Process the TTML <style> element
  '''

  class ParsingContext(TTMLElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

    def __init__(self, parent_ctx: typing.Optional[TTMLElement.ParsingContext] = None):
      self.styles: typing.Dict[model_styles.StyleProperty, typing.Any] = dict()
      self.style_refs: typing.Optional[typing.List[str]] = None
      self.id: typing.Optional[str] = None
      super().__init__(StyleElement, parent_ctx)

  qn = f"{{{xml_ns.TTML}}}style"

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == StyleElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[StyleElement.ParsingContext]:
    '''Converts the XML element `xml_elem` into its representation in the data model.
    `parent_ctx` contains state information passed from parent to child in the TTML hierarchy.
    '''

    style_ctx = StyleElement.ParsingContext(parent_ctx)

    # collect all specified style attributes

    for attr in xml_elem.attrib:
      prop = StyleProperties.BY_QNAME.get(attr)

      if prop is None:
        continue

      try:

        model_prop, model_value = prop.to_model(style_ctx, xml_elem)
        style_ctx.styles[model_prop] = model_value

      except ValueError as e:

        LOGGER.error("Error reading style property %s with %s", prop.__name__, str(e))

    # merge nested style attributes if the parent is a region element

    if issubclass(parent_ctx.ttml_class, RegionElement):

      for style_prop, value in style_ctx.styles.items():
        region_style = parent_ctx.model_element.get_style(style_prop)

        if region_style is None:
          parent_ctx.model_element.set_style(style_prop, value)

      return None

    # process other attributes

    style_ctx.style_refs = imsc_attr.StyleAttribute.extract(xml_elem)

    style_ctx.id = imsc_attr.XMLIDAttribute.extract(xml_elem)

    if style_ctx.id is None:
      LOGGER.error("A style element must have an id")
      return None

    return style_ctx


class InitialElement(TTMLElement):
  '''Process the TTML <initial> element
  '''

  class ParsingContext(TTMLElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}initial"

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == InitialElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[InitialElement.ParsingContext]:
    '''Converts the XML element `xml_elem` into its representation in the data model.
    `parent_ctx` contains state information passed from parent to child in the TTML hierarchy.
    '''

    initial_ctx = InitialElement.ParsingContext(InitialElement, parent_ctx)

    # collect the specified style attributes

    for attr in xml_elem.attrib:

      prop = StyleProperties.BY_QNAME.get(attr)

      if prop is None:

        continue

      try:

        # set the initial value on the data model ContentDocument (the data model does have 
        # a distinct <initial> element)

        model_prop, model_value = prop.to_model(initial_ctx, xml_elem)

        initial_ctx.doc.put_initial_value(model_prop, model_value)

      except (ValueError, TypeError) as e:

        LOGGER.error("Error reading style property %s with %s", prop.__name__, str(e))

    return initial_ctx


  @staticmethod
  def from_model(style_prop: imsc_styles.StyleProperty, initial_value: typing.Any):
    '''Returns a TTML `initial` element corresponding to the style property `style_prop` with
    initial value `initial_value`.
    '''

    initial_element = et.Element(InitialElement.qn)
    
    style_prop.from_model(initial_element, initial_value)

    return initial_element


#
# process content elements
#

class ContentElement(TTMLElement):
  '''TTML content elements: body, div, p, span, br
  '''

  class ParsingContext(TTMLElement.ParsingContext):
    '''Maintains state when parsing the element
    '''
    def __init__(
        self,
        ttml_class: typing.Optional[typing.Type[ContentElement]],
        parent_ctx: TTMLElement.ParsingContext,
        model_element: typing.Optional[model.ContentElement] = None
      ):
      self.children: typing.List[model.ContentElement] = []
      self.model_element: model.ContentElement = model_element
      super().__init__(ttml_class, parent_ctx)

    def process_region_property(self, xml_elem):
      '''Reads and processes the `region` attribute
      '''
      rid = imsc_attr.RegionAttribute.extract(xml_elem)

      if rid is None:
        return

      r = self.doc.get_region(rid)
      
      if r is not None:
        self.model_element.set_region(r)
      else:
        LOGGER.warning("Element references unknown region")

    def process_referential_styling(self, xml_elem):
      '''Processes referential styling
      '''
      for style_ref in reversed(imsc_attr.StyleAttribute.extract(xml_elem)):
        style_element = self.style_elements.get(style_ref)

        if style_element is None:
          LOGGER.error("non existant style id")
          continue

        for model_prop, value in style_element.styles.items():
          if not self.model_element.has_style(model_prop):
            self.model_element.set_style(model_prop, value)

    def process_specified_styling(self, xml_elem):
      '''Processes specified styling
      '''
      for attr in xml_elem.attrib:
        prop = StyleProperties.BY_QNAME.get(attr)

        if prop is None:
          continue

        try:
          model_prop, model_value = prop.to_model(self, xml_elem)

          self.model_element.set_style(model_prop, model_value)

        except ValueError as e:

          LOGGER.error("Error reading style property %s with %s", prop.__name__, str(e))

    def process_set_style_properties(self, parent_ctx: ContentElement.ParsingContext, xml_elem):
      '''Processes style properties on `<set>` element
      '''
      if parent_ctx.model_element is None:
        LOGGER.error("Set parent does not exist")
        return

      if not issubclass(parent_ctx.ttml_class, ContentElement):
        LOGGER.error("Set parent is not a content element")
        return

      for attr in xml_elem.attrib:
        prop = StyleProperties.BY_QNAME.get(attr)
        if prop is not None:
          try:
            model_prop, model_value = prop.to_model(self, xml_elem)
            parent_ctx.model_element.add_animation_step(
              model.DiscreteAnimationStep(
                model_prop,
                self.desired_begin,
                self.desired_end,
                model_value
              )
            )
            break
          except ValueError as e:
            LOGGER.error("Error reading style property %s with %s", prop.__name__, str(e))

    def process_lang_attribute(self, parent_ctx: TTMLElement.ParsingContext, xml_elem):
      super().process_lang_attribute(parent_ctx, xml_elem)
      self.model_element.set_lang(self.lang)

    def process_space_attribute(self, parent_ctx: TTMLElement.ParsingContext, xml_elem):
      super().process_space_attribute(parent_ctx, xml_elem)
      self.model_element.set_space(self.space)

    # pylint: disable=too-many-branches

    def process(self, parent_ctx: TTMLElement.ParsingContext, xml_elem: et.Element):
      '''Generic processing applicable to TTML elements rooted in `region` and `body` elements
      '''
      self.process_lang_attribute(parent_ctx, xml_elem)

      self.process_space_attribute(parent_ctx, xml_elem)

      if self.ttml_class.has_region:
        self.process_region_property(xml_elem)

      # temporal processing. Sequential time containers are converted to parallel time containers since the data model does not
      # support the former.

      self.time_container = imsc_attr.TimeContainerAttribute.extract(xml_elem)

      self.explicit_begin = imsc_attr.BeginAttribute.extract(self.temporal_context, xml_elem)

      self.explicit_dur = imsc_attr.DurAttribute.extract(self.temporal_context, xml_elem)

      self.explicit_end = imsc_attr.EndAttribute.extract(self.temporal_context, xml_elem)

      if parent_ctx.time_container.is_par():
        self.implicit_begin = Fraction(0)
      else:      
        self.implicit_begin = parent_ctx.implicit_end - parent_ctx.desired_begin
      
      self.desired_begin = self.implicit_begin + (self.explicit_begin if self.explicit_begin is not None else Fraction(0))

      if issubclass(self.ttml_class, (BrElement, RegionElement, SetElement)) and \
        parent_ctx.time_container.is_par():
        # br, region and set elements have indefinite duration in parallel time containers

        self.implicit_end = None
      else:
        self.implicit_end = self.desired_begin
        
      # process text nodes

      if self.ttml_class.is_mixed and xml_elem.text is not None and self.time_container.is_par():
        self.children.append(ContentElement.make_anonymous_span(self.doc, self.model_element, xml_elem.text))
        self.implicit_end = None

      # process children elements

      is_inline_animation_complete = False

      for child_xml_element in xml_elem:

        if issubclass(self.ttml_class, RegionElement) and StyleElement.is_instance(child_xml_element):
          # process nest styling, which is specific to region elements, and does not affect temporal
          # processing
          StyleElement.from_xml(self, child_xml_element)
          continue

        child_element = ContentElement.from_xml(self, child_xml_element)

        if child_element is not None:

          if issubclass(child_element.ttml_class, SetElement):
            if is_inline_animation_complete:
              LOGGER.warning("<set> element is out of order")
          elif is_inline_animation_complete is False:
            is_inline_animation_complete = True

          if self.time_container.is_seq():

            self.implicit_end = None if child_element.desired_end is None else child_element.desired_end + self.desired_begin

          else:

            if self.implicit_end is not None and child_element.desired_end is not None:

              self.implicit_end = max(self.implicit_end, child_element.desired_end)

            else:

              self.implicit_end = None

          # skip child if it has no temporal extent

          if not issubclass(child_element.ttml_class, SetElement) and \
            (child_element.desired_begin is None or child_element.desired_end is None or \
              child_element.desired_begin != child_element.desired_end):

            self.children.append(child_element.model_element)

        # process tail text node

        if self.ttml_class.is_mixed and child_xml_element.tail is not None and self.time_container.is_par():
          self.children.append(
            ContentElement.make_anonymous_span(
              self.doc,
              self.model_element,
              child_xml_element.tail
              )
            )
          self.implicit_end = None

      # process referential styling last since it has the lowest priority compared to specified and nested styling

      if self.ttml_class.has_styles:

        self.process_referential_styling(xml_elem)

      try:
        
        if self.ttml_class.has_children:
          self.model_element.push_children(self.children)

      except (ValueError, TypeError) as e:

        LOGGER.error(str(e))

        return

      # temporal end processing

      if self.explicit_end is not None and self.explicit_dur is not None:

        self.desired_end = min(self.desired_begin + self.explicit_dur, self.implicit_begin + self.explicit_end)
        self.desired_end = min(self.desired_begin + self.explicit_dur, self.implicit_begin + self.explicit_end)

      elif self.explicit_end is None and self.explicit_dur is not None:

        self.desired_end = self.desired_begin + self.explicit_dur

      elif self.explicit_end is not None and self.explicit_dur is None:

        self.desired_end = self.implicit_begin + self.explicit_end

      else:

        self.desired_end = self.implicit_end

      if self.ttml_class.has_timing:

        # temporal processing applies to all contents elements but explicit begin and end properties are stored only if their values
        # are not constant, e.g. br elements always have indefinit begin and end times in parallel time containers

        self.model_element.set_begin(self.desired_begin if self.desired_begin != 0 else None)

        self.model_element.set_end(self.desired_end)

      # process style properties

      if issubclass(self.ttml_class, SetElement):

        self.process_set_style_properties(parent_ctx, xml_elem)

      elif self.ttml_class.has_styles:

        self.process_specified_styling(xml_elem)

    # pylint: enable=too-many-branches

  @classmethod
  def make_ttml_element(cls):
    '''Creates an XML element for the content element
    '''
    raise NotImplementedError

  @staticmethod
  def from_model_style_properties(model_content_element, element):
    '''Write TTML style properties from the model'''

    for model_prop, imsc_prop in StyleProperties.BY_MODEL_PROP.items():
      value = model_content_element.get_style(model_prop)
      if value is not None:
        imsc_prop.from_model(element, value)

  @staticmethod
  def from_model_animation(ctx: TTMLElement.WritingContext, model_element: model.ContentElement, xml_element):
    '''Write TTML set element from the model'''

    for a_step in model_element.iter_animation_steps():
      set_element = SetElement.from_model(ctx, a_step)
      if set_element is not None:
        xml_element.append(set_element)


  @property
  def has_timing(self):
    '''`True` if the element supports temporal attributes
    '''
    raise NotImplementedError

  @property
  def has_region(self):
    '''`True` if the element can reference a region
    '''
    raise NotImplementedError

  @property
  def has_styles(self):
    '''`True` if the element can contain style properties
    '''
    raise NotImplementedError

  @property
  def is_mixed(self):
    '''`True` if the element can contain text
    '''
    raise NotImplementedError

  @property
  def has_children(self):
    '''`True` if the element can contain children elements
    '''
    raise NotImplementedError

  @staticmethod
  def make_anonymous_span(document, model_element, span_text):
    '''Creates an anonymous span in the element `model_element` from the text contained in `span_text`
    '''
    if isinstance(model_element, model.Span):
      return model.Text(document, span_text)

    s = model.Span(document)
    s.set_space(model_element.get_space())
    s.set_lang(model_element.get_lang())

    s.push_child(model.Text(document, span_text))

    return s

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[ContentElement.ParsingContext]:
    '''Converts the XML element `xml_elem` into its representation in the data model.
    `parent_ctx` contains state information passed from parent to child in the TTML hierarchy.
    '''

    content_classes = [
      BodyElement,
      DivElement,
      PElement,
      SpanElement,
      RubyElement,
      RbElement,
      RtElement,
      RpElement,
      RbcElement,
      RtcElement,
      BrElement,
      SetElement,
      RegionElement
      ]

    for ttml_elem_class in content_classes:
      if ttml_elem_class.is_instance(xml_elem):
        return ttml_elem_class.from_xml(parent_ctx, xml_elem)
    
    return None

  # pylint: disable=too-many-branches

  @staticmethod
  def from_model(
    ctx: TTMLElement.WritingContext,
    model_element: model.ContentElement
  ) -> typing.Optional[et.Element]:
    '''Returns the TTML element corresponding to the model element `model_element`.
    `ctx` contains state information used in the process.
    '''

    if isinstance(model_element, model.Body):
      imsc_class = BodyElement
    elif isinstance(model_element, model.Div):
      imsc_class = DivElement
    elif isinstance(model_element, model.P):
      imsc_class = PElement
    elif isinstance(model_element, model.Span):
      imsc_class = SpanElement
    elif isinstance(model_element, model.Br):
      imsc_class = BrElement
    elif isinstance(model_element, model.Ruby):
      imsc_class = RubyElement
    elif isinstance(model_element, model.Rb):
      imsc_class = RbElement
    elif isinstance(model_element, model.Rt):
      imsc_class = RtElement
    elif isinstance(model_element, model.Rbc):
      imsc_class = RbcElement
    elif isinstance(model_element, model.Rtc):
      imsc_class = RtcElement
    elif isinstance(model_element, model.Region):
      imsc_class = RegionElement
    else:
      return None

    xml_element = imsc_class.make_ttml_element()

    if (model_element.parent() is None and model_element.get_space() is model.WhiteSpaceHandling.PRESERVE) or \
      (model_element.parent() is not None and model_element.parent().get_space() != model_element.get_space()):
      imsc_attr.XMLSpaceAttribute.set(xml_element, model_element.get_space())

    if imsc_class.has_region:
      if model_element.get_region() is not None:
        imsc_attr.RegionAttribute.set(xml_element, model_element.get_region().get_id())

    if imsc_class.has_timing:
      if model_element.get_begin() is not None:
        imsc_attr.BeginAttribute.set(ctx.temporal_context, xml_element, model_element.get_begin())

      if model_element.get_end() is not None:
        imsc_attr.EndAttribute.set(ctx.temporal_context, xml_element, model_element.get_end())

    if model_element.get_id() is not None:
      imsc_attr.XMLIDAttribute.set(xml_element, model_element.get_id())

    if imsc_class.has_styles:
      ContentElement.from_model_style_properties(model_element, xml_element)
      ContentElement.from_model_animation(ctx, model_element, xml_element)

    if imsc_class.has_children:
      last_child_element = None

      for child in iter(model_element):
        if isinstance(child, model.Text):
          if last_child_element is None:
            xml_element.text = child.get_text()
          else:
            last_child_element.tail = child.get_text()

        child_element = ContentElement.from_model(ctx, child)
        if child_element is not None:
          xml_element.append(child_element)
        
        last_child_element = child_element

    return xml_element

class RegionElement(ContentElement):
  '''Process the TTML <region> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}region"
  has_timing = True
  has_region = False
  has_styles = True
  is_mixed = False
  has_children = False

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == RegionElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RegionElement.ParsingContext]:
    rid = imsc_attr.XMLIDAttribute.extract(xml_elem)

    if rid is None:
      LOGGER.error("All regions must have an id")
      return None

    region_ctx = RegionElement.ParsingContext(RegionElement, parent_ctx, model.Region(rid, parent_ctx.doc))
    region_ctx.process(parent_ctx, xml_elem)
    return region_ctx

  @staticmethod
  def from_model(
    ctx: TTMLElement.WritingContext,
    model_element: model.Region
  ) -> typing.Optional[et.Element]:
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)

class SetElement(ContentElement):
  '''Process TTML <set> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''
    def process_lang_attribute(self, parent_ctx: TTMLElement.ParsingContext, xml_elem):
      # <set> ignores xml:lang
      pass

    def process_space_attribute(self, parent_ctx: TTMLElement.ParsingContext, xml_elem):
      # <set> ignores xml:space
      pass

  qn = f"{{{xml_ns.TTML}}}set"
  has_region = False
  has_styles = False
  has_timing = False
  is_mixed = False
  has_children = False

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == SetElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[SetElement.ParsingContext]:
    set_ctx = SetElement.ParsingContext(SetElement, parent_ctx)
    set_ctx.process(parent_ctx, xml_elem)

    return set_ctx

  @staticmethod
  def from_model(
    ctx: TTMLElement.WritingContext,
    model_element: model.DiscreteAnimationStep
  ) -> typing.Optional[et.Element]:

    set_element = et.Element(SetElement.qn)

    imsc_style = imsc_styles.StyleProperties.BY_MODEL_PROP[model_element.style_property]

    imsc_style.from_model(
      set_element,
      model_element.value
    )

    if model_element.begin is not None:
      imsc_attr.BeginAttribute.set(ctx.temporal_context, set_element, model_element.begin)

    if model_element.end is not None:
      imsc_attr.EndAttribute.set(ctx.temporal_context, set_element, model_element.end)

    return set_element

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)


class BodyElement(ContentElement):
  '''Process TTML body element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}body"
  has_region = True
  has_styles = True
  has_timing = True
  is_mixed = False
  has_children = True

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == BodyElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[BodyElement.ParsingContext]:
    body_ctx = BodyElement.ParsingContext(BodyElement, parent_ctx, model.Body(parent_ctx.doc))
    body_ctx.process(parent_ctx, xml_elem)
    return body_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)

class DivElement(ContentElement):
  '''Process TTML <div> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}div"
  has_region = True
  has_styles = True
  has_timing = True
  is_mixed = False
  has_children = True

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == DivElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[DivElement.ParsingContext]:
    div_ctx = DivElement.ParsingContext(DivElement, parent_ctx, model.Div(parent_ctx.doc))
    div_ctx.process(parent_ctx, xml_elem)
    return div_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)

class PElement(ContentElement):
  '''Process TTML <p> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}p"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = True
  has_children = True

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == PElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[PElement.ParsingContext]:
    p_ctx = PElement.ParsingContext(PElement, parent_ctx, model.P(parent_ctx.doc))
    p_ctx.process(parent_ctx, xml_elem)
    return p_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)

class SpanElement(ContentElement):
  '''Process the TTML <span> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}span"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = True
  has_children = True

  ruby_attribute_qn = f"{{{xml_ns.TTS}}}ruby"

  @staticmethod
  def is_instance(xml_elem):
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) is None

  @staticmethod
  def get_ruby_attr(ttml_span):
    '''extracts the value of the TTML `tts:ruby` attribute from the XML element `ttml_span`
    '''
    return ttml_span.get(SpanElement.ruby_attribute_qn)

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[SpanElement.ParsingContext]:
    span_ctx = SpanElement.ParsingContext(SpanElement, parent_ctx, model.Span(parent_ctx.doc))
    span_ctx.process(parent_ctx, xml_elem)
    return span_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)

class RubyElement(ContentElement):
  '''Process the TTML <span tts:ruby="container"> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}span"
  ruby = "container"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = False
  has_children = True

  @staticmethod
  def is_instance(xml_elem):
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) == RubyElement.ruby

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RubyElement.ParsingContext]:
    ruby_ctx = RubyElement.ParsingContext(RubyElement, parent_ctx, model.Ruby(parent_ctx.doc))
    ruby_ctx.process(parent_ctx, xml_elem)
    return ruby_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn, {SpanElement.ruby_attribute_qn: cls.ruby})

class RbElement(ContentElement):
  '''Process the TTML <span tts:ruby="base"> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  ruby = "base"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = True
  has_children = True

  @staticmethod
  def is_instance(xml_elem):
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) == RbElement.ruby

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RbElement.ParsingContext]:
    rb_ctx = RbElement.ParsingContext(RbElement, parent_ctx, model.Rb(parent_ctx.doc))
    rb_ctx.process(parent_ctx, xml_elem)
    return rb_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(RubyElement.qn, {SpanElement.ruby_attribute_qn: cls.ruby})


class RtElement(ContentElement):
  '''Process the TTML <span tts:ruby="text"> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  ruby = "text"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = True
  has_children = True

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) == RtElement.ruby

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RtElement.ParsingContext]:
    rt_ctx = RtElement.ParsingContext(RtElement, parent_ctx, model.Rt(parent_ctx.doc))
    rt_ctx.process(parent_ctx, xml_elem)
    return rt_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(RubyElement.qn, {SpanElement.ruby_attribute_qn: cls.ruby})


class RpElement(ContentElement):
  '''Process the TTML <span tts:ruby="delimiter"> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  ruby = "delimiter"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = True
  has_children = True

  @staticmethod
  def is_instance(xml_elem):
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) == RpElement.ruby

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RpElement.ParsingContext]:
    rp_ctx = RpElement.ParsingContext(RpElement, parent_ctx, model.Rp(parent_ctx.doc))
    rp_ctx.process(parent_ctx, xml_elem)
    return rp_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(RubyElement.qn, {SpanElement.ruby_attribute_qn: cls.ruby})


class RbcElement(ContentElement):
  '''Process the TTML <span tts:ruby="baseContainer"> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  ruby = "baseContainer"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = False
  has_children = True

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) == RbcElement.ruby

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RbcElement.ParsingContext]:
    rbc_ctx = RbcElement.ParsingContext(RbcElement, parent_ctx, model.Rbc(parent_ctx.doc))
    rbc_ctx.process(parent_ctx, xml_elem)
    return rbc_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(RubyElement.qn, {SpanElement.ruby_attribute_qn: cls.ruby})


class RtcElement(ContentElement):
  '''Process the TTML <span tts:ruby="textContainer"> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  ruby = "textContainer"
  has_timing = True
  has_region = True
  has_styles = True
  is_mixed = False
  has_children = True

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == SpanElement.qn and SpanElement.get_ruby_attr(xml_elem) == RtcElement.ruby

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[RtcElement.ParsingContext]:
    rtc_ctx = RtcElement.ParsingContext(RtcElement, parent_ctx, model.Rtc(parent_ctx.doc))
    rtc_ctx.process(parent_ctx, xml_elem)
    return rtc_ctx

  @staticmethod
  def from_model(ctx: TTMLElement.WritingContext, model_element: model.ContentElement):
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(RubyElement.qn, {SpanElement.ruby_attribute_qn: cls.ruby})


class BrElement(ContentElement):
  '''Process the TTML <br> element
  '''

  class ParsingContext(ContentElement.ParsingContext):
    '''Maintains state when parsing the element
    '''

  qn = f"{{{xml_ns.TTML}}}br"
  has_timing = False
  has_region = False
  has_styles = True
  is_mixed = False
  has_children = False

  @staticmethod
  def is_instance(xml_elem) -> bool:
    return xml_elem.tag == BrElement.qn

  @staticmethod
  def from_xml(
    parent_ctx: typing.Optional[TTMLElement.ParsingContext],
    xml_elem: et.Element
  ) -> typing.Optional[BrElement.ParsingContext]:
    br_ctx = BrElement.ParsingContext(BrElement, parent_ctx, model.Br(parent_ctx.doc))
    br_ctx.process(parent_ctx, xml_elem)
    return br_ctx

  @staticmethod
  def from_model(
    ctx: TTMLElement.WritingContext,
    model_element: typing.Any
  ) -> typing.Optional[et.Element]:
    return ContentElement.from_model(ctx, model_element)

  @classmethod
  def make_ttml_element(cls):
    return et.Element(cls.qn)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the data model'''

# pylint: disable=R0201,C0115,C0116

import typing
import unittest
from fractions import Fraction
import ttconv.model as model
import ttconv.style_properties as styles

class ContentDocumentTest(unittest.TestCase):

  def test_init(self):
    d = model.ContentDocument()

    self.assertIsNone(d.get_body())
    
    self.assertEqual(d.get_px_resolution(), model.PixelResolutionType(width=1920, height=1080))

    self.assertEqual(len(d.iter_initial_values()), 0)

    self.assertEqual(len(d.iter_regions()), 0)

    self.assertEqual(d.get_cell_resolution(), model.CellResolutionType(rows=15, columns=32))
  
  
  def test_set_body(self):
    d = model.ContentDocument()

    b = model.Body(d)

    d.set_body(b)

    self.assertEqual(d.get_body(), b)

    d.set_body(None)

    self.assertIsNone(d.get_body())

  def test_set_bad_body(self):
    d = model.ContentDocument()

    b = model.Body()

    with self.assertRaises(ValueError):
      d.set_body(b)
  
  def test_add_region(self):
    d = model.ContentDocument()

    r = model.Region("hello", d)

    d.put_region(r)

    self.assertIs(d.get_region(r.get_id()), r)

  def test_remove_region(self):
    d = model.ContentDocument()

    r = model.Region("hello", d)

    d.put_region(r)

    self.assertIs(d.get_region(r.get_id()), r)

    d.remove_region(r.get_id())

    self.assertFalse(d.has_region(r.get_id()))

    self.assertIsNone(d.get_region(r.get_id()))


  def test_add_detached_region(self):
    d = model.ContentDocument()

    r = model.Region("hello")

    with self.assertRaises(ValueError):
      d.put_region(r)


  def test_add_dup_region(self):
    d = model.ContentDocument()

    r1 = model.Region("hello", d)

    r2 = model.Region("hello", d)

    d.put_region(r1)

    d.put_region(r2)

    self.assertIs(d.get_region(r2.get_id()), r2)

  def test_iter_region(self):
    d = model.ContentDocument()

    r1 = model.Region("hello1", d)

    r2 = model.Region("hello2", d)

    d.put_region(r1)

    d.put_region(r2)

    self.assertCountEqual(d.iter_regions(), [r1, r2])


  def test_add_initial_value(self):
    d = model.ContentDocument()

    c = styles.ColorType((12, 23, 43, 56))

    d.put_initial_value(styles.StyleProperties.Color, c)

    c2 = d.get_initial_value(styles.StyleProperties.Color)

    self.assertEqual(c, c2)

  def test_remove_initial_value(self):
    d = model.ContentDocument()

    c = styles.ColorType((12, 23, 43, 56))

    d.put_initial_value(styles.StyleProperties.Color, c)

    c2 = d.get_initial_value(styles.StyleProperties.Color)

    self.assertEqual(c, c2)

    d.remove_initial_value(styles.StyleProperties.Color)

    self.assertFalse(d.has_initial_value(styles.StyleProperties.Color))

    self.assertIsNone(d.get_initial_value(styles.StyleProperties.Color))

  def test_add_null_initial_value(self):
    d = model.ContentDocument()

    c = styles.ColorType((12, 23, 43, 56))

    d.put_initial_value(styles.StyleProperties.Color, c)

    c2 = d.get_initial_value(styles.StyleProperties.Color)

    self.assertEqual(c, c2)

    d.put_initial_value(styles.StyleProperties.Color, None)

    self.assertFalse(d.has_initial_value(styles.StyleProperties.Color))

    self.assertIsNone(d.get_initial_value(styles.StyleProperties.Color))

  def test_add_dup_initial_value(self):
    d = model.ContentDocument()

    c1 = styles.ColorType((12, 23, 43, 56))

    c2 = styles.ColorType((12, 96, 43, 56))

    d.put_initial_value(styles.StyleProperties.Color, c1)

    d.put_initial_value(styles.StyleProperties.Color, c2)

    self.assertIsNot(d.get_initial_value(styles.StyleProperties.Color), c1)

    self.assertIs(d.get_initial_value(styles.StyleProperties.Color), c2)

  def test_add_bad_initial_value(self):
    d = model.ContentDocument()

    c1 = styles.ColorType((12, 23, 43, 56))
    
    with self.assertRaises(ValueError):
      d.put_initial_value(styles.StyleProperties.Extent, c1)

  def test_iter_initial_values(self):
    d = model.ContentDocument()

    c1 = styles.ColorType((12, 23, 43, 56))

    c2 = styles.ColorType((12, 96, 43, 56))

    d.put_initial_value(styles.StyleProperties.Color, c1)

    d.put_initial_value(styles.StyleProperties.BackgroundColor, c2)

    self.assertCountEqual(
      [(styles.StyleProperties.Color, c1), (styles.StyleProperties.BackgroundColor, c2)],
      d.iter_initial_values()
    )

  def test_cell_resolution(self):

    cr = model.CellResolutionType(rows=10, columns=20)

    self.assertEqual(cr.columns, 20)

    self.assertEqual(cr.rows, 10)

    with self.assertRaises(ValueError):
      cr = model.CellResolutionType(rows=0, columns=20)

  def test_content_profiles(self):
    d = model.ContentDocument()

    cp : typing.Set[str] = {"http://www.w3.org/ns/ttml/profile/imsc1.1/text", "http://www.w3.org/ns/ttml/profile/imsc1/text"}

    d.set_content_profiles(cp)

    self.assertSetEqual(cp, d.get_content_profiles())

    d.set_content_profiles(None)

    self.assertIsNone(d.get_content_profiles())

    with self.assertRaises(TypeError):
      d.set_content_profiles({"b a", "http://www.w3.org/ns/ttml/profile/imsc1.1/text"})

    with self.assertRaises(TypeError):
      d.set_content_profiles({1, "http://www.w3.org/ns/ttml/profile/imsc1.1/text"})


  def test_language(self):

    d = model.ContentDocument()

    self.assertEqual(d.get_lang(), "")

    lang = "fr-CA"

    d.set_lang(lang)

    self.assertEqual(d.get_lang(), lang)

    with self.assertRaises(TypeError):
      d.set_lang(1)

  def test_pixel_resolution(self):

    cr = model.PixelResolutionType(height=480, width=640)

    self.assertEqual(cr.height, 480)

    self.assertEqual(cr.width, 640)

    with self.assertRaises(ValueError):
      cr = model.PixelResolutionType(height=0, width=640)


  def test_active_area_default(self):

    aa = model.ActiveAreaType()

    self.assertEqual(aa.left_offset, 0)

    self.assertEqual(aa.top_offset, 0)

    self.assertEqual(aa.height, 1)

    self.assertEqual(aa.width, 1)


  def test_active_area(self):

    aa = model.ActiveAreaType(0.1, 0.15, 0.8, 0.7)

    self.assertEqual(aa.left_offset, 0.1)

    self.assertEqual(aa.top_offset, 0.15)

    self.assertEqual(aa.width, 0.8)

    self.assertEqual(aa.height, 0.7)

    with self.assertRaises(ValueError):
      model.ActiveAreaType(-0.1, 0.15, 0.8, 0.7)

    with self.assertRaises(ValueError):
      model.ActiveAreaType(0.1, 0.15, -0.8, 0.7)


  def test_document_active_area_default(self):

    d = model.ContentDocument()

    self.assertIsNone(d.get_active_area())


  def test_document_active_area(self):

    d = model.ContentDocument()

    aa = model.ActiveAreaType(0.1, 0.15, 0.8, 0.7)

    d.set_active_area(aa)

    self.assertEqual(
      d.get_active_area(),
      aa
      )

    d.set_active_area(None)

    self.assertIsNone(d.get_active_area())


  def test_display_aspect_ratio(self):

    d = model.ContentDocument()

    self.assertIsNone(d.get_display_aspect_ratio())

    d.set_display_aspect_ratio(Fraction(16, 9))

    self.assertEqual(d.get_display_aspect_ratio(), Fraction(16, 9))

    d.set_display_aspect_ratio(None)

    self.assertIsNone(d.get_display_aspect_ratio())

    with self.assertRaises(TypeError):
      d.set_display_aspect_ratio("hello")

  def test_copy_to(self):
    src = model.ContentDocument()

    src.set_display_aspect_ratio(Fraction(16, 9))
    src.set_active_area(model.ActiveAreaType(0.1, 0.15, 0.8, 0.7))
    src.set_px_resolution(model.PixelResolutionType(height=480, width=640))
    src.set_lang("fr")
    src.set_cell_resolution(model.CellResolutionType(rows=10, columns=20))

    src.put_initial_value(styles.StyleProperties.Color, styles.ColorType((12, 23, 43, 56)))

    dest = model.ContentDocument()

    src.copy_to(dest)

    self.assertEqual(dest.get_display_aspect_ratio(), src.get_display_aspect_ratio())
    self.assertEqual(dest.get_active_area(), src.get_active_area())
    self.assertEqual(dest.get_px_resolution(), src.get_px_resolution())
    self.assertEqual(dest.get_lang(), src.get_lang())
    self.assertEqual(dest.get_cell_resolution(), src.get_cell_resolution())

    self.assertSequenceEqual(list(dest.iter_initial_values()), list(src.iter_initial_values()))

if __name__ == '__main__':
  unittest.main()



"""A setuptools based setup module.
"""

import pathlib
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='ttconv',
    version='1.2.2',
    description='Library for conversion of common timed text formats',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sandflow/ttconv',
    author='Sandflow Consulting LLC',
    author_email='info@sandflow.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Environment :: Console',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Multimedia'
    ],
    keywords='ttml, timed text, caption, subtitle, imsc, scc, srt, stl, smpte-tt, conversion, vtt, webvtt, 608',

    package_dir={'ttconv': 'src/main/python/ttconv'},

    packages=find_packages(where='src/main/python'),

    python_requires='>=3.7, <4',

    project_urls={
        'Bug Reports': 'https://github.com/sandflow/ttconv/issues',
        'Source': 'https://github.com/sandflow/ttconv',
    },

    entry_points={
        "console_scripts": [
            "tt = ttconv.tt:main"
        ]
    },
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for logging and progress bar'''

import logging
import unittest
import ttconv.tt as tt

LOGGER = logging.getLogger("ttconv")

# pylint: disable=R0201,C0115,C0116

class IMSCAppLoggingProgressBarTest(unittest.TestCase):

  def test_logging_info(self):

    with self.assertLogs('ttconv', level='INFO') as cm:
      inputfile = "input"
      outputfile = "output"
      LOGGER.info("Input file is %s", inputfile)
      LOGGER.info("Output file is %s", outputfile)
      LOGGER.info('test 1')

      LOGGER.info('test 2')
      LOGGER.info('test 3')
      LOGGER.info('test 4')
      LOGGER.info('test 5')
      LOGGER.info('test 6')
      LOGGER.info('test 7')

    self.assertEqual(cm.output, [
                                'INFO:ttconv:Input file is input',
                                'INFO:ttconv:Output file is output',
                                'INFO:ttconv:test 1',
                                'INFO:ttconv:test 2',
                                'INFO:ttconv:test 3',
                                'INFO:ttconv:test 4',
                                'INFO:ttconv:test 5',
                                'INFO:ttconv:test 6',
                                'INFO:ttconv:test 7'
                                ])

  def test_logging_progress_bar(self):

    inputfile = "input"
    outputfile = "output"
    LOGGER.info("Input file is %s", inputfile)
    LOGGER.info("Output file is %s", outputfile)
    LOGGER.info('test 1')

    tt.progress_callback_read(0)
    LOGGER.info('test 2')
    LOGGER.info('test 3')
    tt.progress_callback_read(0.1)
    tt.progress_callback_read(1)
    LOGGER.info('test 4')
    tt.progress_callback_write(0.3)
    LOGGER.info('test 5')
    tt.LOGGER.info('test 6')
    tt.progress_callback_write(1)
    LOGGER.info('test 7')

  def test_logging_progress_bar_read_write_interlaced(self):
    
    LOGGER.info('Interlaced read/write')
    tt.progress_callback_read(0)
    tt.progress_callback_write(0)
    tt.progress_callback_read(0.1)
    tt.progress_callback_write(0.3)
    tt.progress_callback_write(1)
    tt.progress_callback_read(1)

    LOGGER.info('read then write')
    tt.progress_callback_read(0)
    tt.progress_callback_read(0.1)
    tt.progress_callback_read(1)
    tt.progress_callback_write(0)
    tt.progress_callback_write(0.3)
    tt.progress_callback_write(1)

  def test_logging_progress_bar_asserts(self):

    with self.assertLogs('ttconv', level='INFO') as cm:
      inputfile = "input"
      outputfile = "output"
      LOGGER.info("Input file is %s", inputfile)
      LOGGER.info("Output file is %s", outputfile)
      LOGGER.info('test 1')

      #tt.progress_callback_read(0)
      LOGGER.info('test 2')
      LOGGER.info('test 3')
      #tt.progress_callback_read(0.1)
      #tt.progress_callback_read(1)
      LOGGER.info('test 4')
      #tt.progress_callback_read(0.3)
      LOGGER.info('test 5')
      LOGGER.info('test 6')
      #tt.progress_callback_read(1)
      LOGGER.info('test 7')

    self.assertEqual(cm.output, [
                                'INFO:ttconv:Input file is input',
                                'INFO:ttconv:Output file is output',
                                'INFO:ttconv:test 1',
                                'INFO:ttconv:test 2',
                                'INFO:ttconv:test 3',
                                'INFO:ttconv:test 4',
                                'INFO:ttconv:test 5',
                                'INFO:ttconv:test 6',
                                'INFO:ttconv:test 7'
                                ])

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the SCC paragraph"""

# pylint: disable=R0201,C0115,C0116,W0212

import unittest
from fractions import Fraction

from ttconv.model import ContentDocument, Span, Br
from ttconv.scc.caption_paragraph import SccCaptionParagraph
from ttconv.scc.caption_style import SccCaptionStyle
from ttconv.scc.caption_line import SccCaptionLine
from ttconv.style_properties import TextAlignType
from ttconv.time_code import SmpteTimeCode, FPS_30


class SccCaptionParagraphTest(unittest.TestCase):

  def test_content(self):
    caption_paragraph = SccCaptionParagraph()
    self.assertEqual(0, caption_paragraph._safe_area_x_offset)
    self.assertEqual(0, caption_paragraph._safe_area_y_offset)

    caption_paragraph = SccCaptionParagraph(4, 2)
    self.assertEqual(4, caption_paragraph._safe_area_x_offset)
    self.assertEqual(2, caption_paragraph._safe_area_y_offset)

    self.assertEqual(caption_paragraph.get_current_line(), caption_paragraph.get_lines()[0])
    self.assertTrue(caption_paragraph.get_current_text().is_empty())
    self.assertEqual(1, len(caption_paragraph._caption_lines))

    caption_paragraph.set_cursor_at(4, 4)
    caption_paragraph.new_caption_text()
    self.assertEqual((4, 4), caption_paragraph.get_cursor())
    self.assertEqual(caption_paragraph.get_current_line(), caption_paragraph.get_lines()[4])
    self.assertEqual(4, caption_paragraph.get_current_line().get_row())
    self.assertEqual(4, caption_paragraph.get_current_line().get_indent())
    self.assertEqual(0, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(0, caption_paragraph.get_current_text().get_cursor())

    caption_paragraph.indent_cursor(3)
    self.assertEqual((4, 7), caption_paragraph.get_cursor())
    self.assertEqual(4, caption_paragraph.get_current_line().get_row())
    self.assertEqual(7, caption_paragraph.get_current_line().get_indent())
    self.assertEqual(0, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(0, caption_paragraph.get_current_text().get_cursor())

    caption_paragraph.append_text("Hello")
    self.assertEqual(5, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(5, caption_paragraph.get_current_text().get_cursor())

    caption_paragraph.set_cursor_at(4, 10)
    self.assertEqual((4, 10), caption_paragraph.get_cursor())
    self.assertEqual(4, caption_paragraph.get_current_line().get_row())
    self.assertEqual(7, caption_paragraph.get_current_line().get_indent())

    self.assertEqual(3, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(3, caption_paragraph.get_current_text().get_cursor())

    caption_paragraph.indent_cursor(2)
    self.assertEqual((4, 12), caption_paragraph.get_cursor())
    self.assertEqual(4, caption_paragraph.get_current_line().get_row())
    self.assertEqual(7, caption_paragraph.get_current_line().get_indent())

    self.assertEqual(5, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(5, caption_paragraph.get_current_text().get_cursor())

    self.assertListEqual([], caption_paragraph.get_last_caption_lines(0))
    self.assertListEqual([caption_paragraph.get_current_line()], caption_paragraph.get_last_caption_lines(1))

    caption_paragraph.set_cursor_at(2, 4)
    caption_paragraph.new_caption_text()
    caption_paragraph.append_text("World")
    self.assertEqual(5, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(5, caption_paragraph.get_current_text().get_cursor())

    self.assertRaisesRegex(RuntimeError, "Cannot roll-Up Unknown-styled caption.",
                           caption_paragraph.roll_up)

    caption_paragraph._caption_style = SccCaptionStyle.PopOn
    self.assertRaisesRegex(RuntimeError, "Cannot roll-Up PopOn-styled caption.",
                           caption_paragraph.roll_up)

    caption_paragraph._caption_style = SccCaptionStyle.RollUp
    caption_paragraph.roll_up()
    self.assertEqual(2, len(caption_paragraph.get_lines()))
    self.assertEqual(caption_paragraph.get_current_line(), caption_paragraph.get_lines().get(1))
    self.assertEqual(1, caption_paragraph.get_lines().get(1).get_row())
    self.assertEqual("World", caption_paragraph.get_lines().get(1).get_current_text().get_text())

    self.assertTrue(isinstance(caption_paragraph.get_lines().get(3), SccCaptionLine))
    self.assertEqual(3, caption_paragraph.get_lines().get(3).get_row())
    self.assertEqual("Hello", caption_paragraph.get_lines().get(3).get_current_text().get_text())

    self.assertListEqual([], caption_paragraph.get_last_caption_lines(0))
    self.assertListEqual([caption_paragraph.get_lines().get(3)], caption_paragraph.get_last_caption_lines(1))
    self.assertListEqual([caption_paragraph.get_lines().get(1), caption_paragraph.get_lines().get(3)],
                         caption_paragraph.get_last_caption_lines(2))

    caption_paragraph.set_cursor_at(15, 0)
    caption_paragraph.new_caption_text()
    caption_paragraph.append_text("!!!")
    self.assertEqual(3, caption_paragraph.get_current_line().get_cursor())
    self.assertEqual(3, caption_paragraph.get_current_text().get_cursor())

    caption_paragraph.roll_up()
    self.assertEqual(3, len(caption_paragraph.get_lines()))
    self.assertTrue(isinstance(caption_paragraph.get_lines().get(0), SccCaptionLine))
    self.assertEqual(0, caption_paragraph.get_lines().get(0).get_row())
    self.assertEqual("World", caption_paragraph.get_lines().get(0).get_current_text().get_text())

    self.assertTrue(isinstance(caption_paragraph.get_lines().get(2), SccCaptionLine))
    self.assertEqual(2, caption_paragraph.get_lines().get(2).get_row())
    self.assertEqual("Hello", caption_paragraph.get_lines().get(2).get_current_text().get_text())

    self.assertEqual(caption_paragraph.get_current_line(), caption_paragraph.get_lines().get(14))
    self.assertEqual(14, caption_paragraph.get_lines().get(14).get_row())
    self.assertEqual("!!!", caption_paragraph.get_lines().get(14).get_current_text().get_text())

    self.assertListEqual([], caption_paragraph.get_last_caption_lines(0))
    self.assertListEqual([caption_paragraph.get_lines().get(14)], caption_paragraph.get_last_caption_lines(1))
    self.assertListEqual(
      [caption_paragraph.get_lines().get(2), caption_paragraph.get_lines().get(14)],
      caption_paragraph.get_last_caption_lines(2))
    self.assertListEqual(
      [caption_paragraph.get_lines().get(0), caption_paragraph.get_lines().get(2), caption_paragraph.get_lines().get(14)],
      caption_paragraph.get_last_caption_lines(3))

  def test_paragraph_alignment_detection_left(self):
    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(0, 0)
    caption_paragraph.append_text("0123456789")

    caption_paragraph.set_cursor_at(1, 0)
    caption_paragraph.append_text("012345")

    caption_paragraph.set_cursor_at(2, 0)
    caption_paragraph.append_text("0123")

    self.assertEqual(TextAlignType.start, caption_paragraph.guess_text_alignment())

  def test_paragraph_alignment_detection_center(self):
    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(0, 0)
    caption_paragraph.append_text("0123456789")

    caption_paragraph.set_cursor_at(1, 2)
    caption_paragraph.append_text("012345")

    caption_paragraph.set_cursor_at(2, 3)
    caption_paragraph.append_text("0123")

    self.assertEqual(TextAlignType.center, caption_paragraph.guess_text_alignment())

  def test_paragraph_alignment_detection_right(self):
    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(0, 0)
    caption_paragraph.append_text("0123456789")

    caption_paragraph.set_cursor_at(1, 4)
    caption_paragraph.append_text("012345")

    caption_paragraph.set_cursor_at(2, 6)
    caption_paragraph.append_text("0123")

    self.assertEqual(TextAlignType.end, caption_paragraph.guess_text_alignment())

  def test_paragraph_content_alignment_detection_left(self):
    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(0, 0)
    caption_paragraph.append_text("  0123456789")

    caption_paragraph.set_cursor_at(1, 0)
    caption_paragraph.append_text("  012345")

    caption_paragraph.set_cursor_at(2, 0)
    caption_paragraph.append_text("  0123")

    self.assertEqual(TextAlignType.start, caption_paragraph.guess_text_alignment())

  def test_paragraph_content_alignment_detection_center(self):
    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(0, 2)
    caption_paragraph.append_text("0123456789")

    caption_paragraph.set_cursor_at(1, 2)
    caption_paragraph.append_text("  012345")

    caption_paragraph.set_cursor_at(2, 2)
    caption_paragraph.append_text("   0123")

    self.assertEqual(TextAlignType.center, caption_paragraph.guess_text_alignment())

    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(0, 0)
    caption_paragraph.append_text("  0123456789")

    caption_paragraph.set_cursor_at(1, 2)
    caption_paragraph.append_text("  012345")

    caption_paragraph.set_cursor_at(2, 5)
    caption_paragraph.append_text("0123")

    self.assertEqual(TextAlignType.center, caption_paragraph.guess_text_alignment())

  def test_paragraph_content_alignment_detection_right(self):
    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(10, 0)
    caption_paragraph.append_text("0123456789")

    caption_paragraph.set_cursor_at(11, 0)
    caption_paragraph.append_text("    012345")

    caption_paragraph.set_cursor_at(12, 0)
    caption_paragraph.append_text("      0123")

    self.assertEqual(TextAlignType.end, caption_paragraph.guess_text_alignment())

    caption_paragraph = SccCaptionParagraph()

    caption_paragraph.set_cursor_at(10, 0)
    caption_paragraph.append_text("0123456789")

    caption_paragraph.set_cursor_at(11, 0)
    caption_paragraph.append_text("    012345")

    caption_paragraph.set_cursor_at(12, 4)
    caption_paragraph.append_text("  0123")

    self.assertEqual(TextAlignType.end, caption_paragraph.guess_text_alignment())

  def test_to_paragraph(self):
    caption_paragraph = SccCaptionParagraph()
    doc = ContentDocument()

    self.assertRaisesRegex(TypeError, "Element id must be a valid xml:id string", caption_paragraph.to_paragraph, doc)

    caption_paragraph.set_id("test-id")

    origin = caption_paragraph.get_origin()
    self.assertEqual(0, origin.x.value)
    self.assertEqual(0, origin.y.value)

    extent = caption_paragraph.get_extent()
    self.assertEqual(0, extent.width.value)
    self.assertEqual(0, extent.height.value)

    paragraph = caption_paragraph.to_paragraph(doc)

    self.assertEqual("test-id", paragraph.get_id())
    self.assertEqual(doc, paragraph.get_doc())
    self.assertIsNone(paragraph.get_begin())
    self.assertIsNone(paragraph.get_end())

    children = list(paragraph)
    self.assertEqual(0, len(children))

    caption_paragraph.set_begin(SmpteTimeCode.parse("00:01:02:03", FPS_30))
    caption_paragraph.set_end(SmpteTimeCode.parse("00:02:03:04", FPS_30))

    caption_paragraph.set_cursor_at(1)
    caption_paragraph.new_caption_text()
    caption_paragraph.append_text("Hello")
    caption_paragraph.set_cursor_at(2)
    caption_paragraph.new_caption_text()
    caption_paragraph.append_text("World")

    paragraph = caption_paragraph.to_paragraph(doc)

    self.assertEqual("test-id", paragraph.get_id())
    self.assertEqual(doc, paragraph.get_doc())
    self.assertEqual(Fraction(1863, 30), paragraph.get_begin())
    self.assertEqual(Fraction(3694, 30), paragraph.get_end())

    children = list(paragraph)
    self.assertEqual(3, len(children))

    self.assertIsInstance(children[0], Span)
    self.assertEqual("Hello", list(children[0])[0].get_text())

    self.assertIsInstance(children[1], Br)

    self.assertIsInstance(children[2], Span)
    self.assertEqual("World", list(children[2])[0].get_text())


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC Extended characters"""

from __future__ import annotations

import typing

from ttconv.scc.codes import SccCode


class SccExtendedCharacter(SccCode):
  """SCC Extended character definition"""

  # Spanish extended characters
  LATIN_CAPITAL_LETTER_A_WITH_ACUTE = (0x1220, 0x1A20, '\u00C1')  # Á capital A with acute accent
  LATIN_CAPITAL_LETTER_E_WITH_ACUTE = (0x1221, 0x1A21, '\u00C9')  # É capital E with acute accent
  LATIN_CAPITAL_LETTER_O_WITH_ACUTE = (0x1222, 0x1A22, '\u00D3')  # Ó capital O with acute accent
  LATIN_CAPITAL_LETTER_U_WITH_ACUTE = (0x1223, 0x1A23, '\u00DA')  # Ú capital U with acute accent
  LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS = (0x1224, 0x1A24, '\u00DC')  # Ü capital U with diaeresis or umlaut
  LATIN_SMALL_LETTER_U_WITH_DIAERESIS = (0x1225, 0x1A25, '\u00FC')  # ü small u with diaeresis or umlaut
  LEFT_SINGLE_QUOTATION_MARK = (0x1226, 0x1A26, '\u2018')  # ‘ opening single quote
  INVERTED_EXCLAMATION_MARK = (0x1227, 0x1A27, '\u00A1')  # ¡ inverted exclamation mark

  # Miscellaneous extended characters
  ASTERISK = (0x1228, 0x1A28, '\u002A')  # * Asterisk
  NEUTRAL_SINGLE_QUOTATION_MARK = (0x1229, 0x1A29, '\u0027')  # ' plain single quote
  BOX_DRAWINGS_HEAVY_HORIZONTAL = (0x122A, 0x1A2A, '\u2501')  # — em dash
  COPYRIGHT_SIGN = (0x122B, 0x1A2B, '\u00A9')  # © Copyright
  SERVICE_MARK = (0x122C, 0x1A2C, '\u2120')  # SM Servicemark
  BULLET = (0x122D, 0x1A2D, '\u2022')  # ● round bullet
  LEFT_DOUBLE_QUOTATION_MARK = (0x122E, 0x1A2E, '\u201C')  # “ opening double quotes
  RIGHT_DOUBLE_QUOTATION_MARK = (0x122F, 0x1A2F, '\u201D')  # ” closing double quotes

  # French extended characters
  LATIN_CAPITAL_LETTER_A_WITH_GRAVE = (0x1230, 0x1A30, '\u00C0')  # À capital A with grave accent
  LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX = (0x1231, 0x1A31, '\u00C2')  # Â capital A with circumflex accent
  LATIN_CAPITAL_LETTER_C_WITH_CEDILLA = (0x1232, 0x1A32, '\u00C7')  # Ç capital C with cedilla
  LATIN_CAPITAL_LETTER_E_WITH_GRAVE = (0x1233, 0x1A33, '\u00C8')  # È capital E with grave accent
  LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX = (0x1234, 0x1A34, '\u00CA')  # Ê capital E with circumflex accent
  LATIN_CAPITAL_LETTER_E_WITH_DIAERESIS = (0x1235, 0x1A35, '\u00CB')  # Ë capital E with diaeresis or umlaut mark
  LATIN_SMALL_LETTER_E_WITH_DIAERESIS = (0x1236, 0x1A36, '\u00EB')  # ë small e with diaeresis or umlaut mark
  LATIN_CAPITAL_LETTER_I_WITH_CIRCUMFLEX = (0x1237, 0x1A37, '\u00CE')  # Î capital I with circumflex accent
  LATIN_CAPITAL_LETTER_I_WITH_DIAERESIS = (0x1238, 0x1A38, '\u00CF')  # Ï capital I with diaeresis or umlaut mark
  LATIN_SMALL_LETTER_I_WITH_DIAERESIS = (0x1239, 0x1A39, '\u00EF')  # ï small i with diaeresis or umlaut mark
  LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX = (0x123A, 0x1A3A, '\u00D4')  # Ô capital O with circumflex
  LATIN_CAPITAL_LETTER_U_WITH_GRAVE = (0x123B, 0x1A3B, '\u00D9')  # Ù capital U with grave accent
  LATIN_SMALL_LETTER_U_WITH_GRAVE = (0x123C, 0x1A3C, '\u00F9')  # ù small u with grave accent
  LATIN_CAPITAL_LETTER_U_WITH_CIRCUMFLEX = (0x123D, 0x1A3D, '\u00DB')  # Û capital U with circumflex accent
  LEFT_POINTING_GUILLEMET = (0x123E, 0x1A3E, '\u00AB')  # « opening guillemets
  RIGHT_POINTING_GUILLEMET = (0x123F, 0x1A3F, '\u00BB')  # » closing guillemets

  # Portuguese extended characters
  LATIN_CAPITAL_LETTER_A_WITH_TILDE = (0x1320, 0x1B20, '\u00C3')  # Ã capital A with tilde
  LATIN_SMALL_LETTER_A_WITH_TILDE = (0x1321, 0x1B21, '\u00E3')  # ã small a with tilde
  LATIN_CAPITAL_LETTER_I_WITH_ACUTE = (0x1322, 0x1B22, '\u00CD')  # Í capital I with acute accent
  LATIN_CAPITAL_LETTER_I_WITH_GRAVE = (0x1323, 0x1B23, '\u00CC')  # Ì capital I with grave accent
  LATIN_SMALL_LETTER_I_WITH_GRAVE = (0x1324, 0x1B24, '\u00EC')  # ì small i with grave accent
  LATIN_CAPITAL_LETTER_O_WITH_GRAVE = (0x1325, 0x1B25, '\u00D2')  # Ò capital O with grave accent
  LATIN_SMALL_LETTER_O_WITH_GRAVE = (0x1326, 0x1B26, '\u00F2')  # ò small o with grave accent
  LATIN_CAPITAL_LETTER_O_WITH_TILDE = (0x1327, 0x1B27, '\u00D5')  # Õ capital O with tilde
  LATIN_SMALL_LETTER_O_WITH_TILDE = (0x1328, 0x1B28, '\u00F5')  # õ small o with tilde
  BRACE_OPENING = (0x1329, 0x1B29, '\u007B')  # { opening brace
  BRACE_CLOSING = (0x132A, 0x1B2A, '\u007D')  # } closing brace
  REVERSE_SOLIDUS = (0x132B, 0x1B2B, '\u005C')  # \ backslash
  LATIN_SMALL_LETTER_TURNED_V = (0x132C, 0x1B2C, '\u028C')  # ^ caret
  LOW_LINE = (0x132D, 0x1B2D, '\u005F')  # _ Underbar
  VERTICAL_LINE = (0x132E, 0x1B2E, '\u007C')  # | pipe
  TILDE = (0x132F, 0x1B2F, '\u007E')  # ~ tilde

  # German extended characters
  LATIN_CAPITAL_LETTER_A_WITH_DIAERESIS = (0x1330, 0x1B30, '\u00C4')  # Ä Capital A with diaeresis or umlaut mark
  LATIN_SMALL_LETTER_A_WITH_DIAERESIS = (0x1331, 0x1B31, '\u00E4')  # ä small a with diaeresis or umlaut mark
  LATIN_CAPITAL_LETTER_O_WITH_DIAERESIS = (0x1332, 0x1B32, '\u00D6')  # Ö Capital O with diaeresis or umlaut mark
  LATIN_SMALL_LETTER_O_WITH_DIAERESIS = (0x1333, 0x1B33, '\u00F6')  # ö small o with diaeresis or umlaut mark
  ESZETT = (0x1334, 0x1B34, '\u00DF')  # ß eszett (mall sharp s)
  YEN_SIGN = (0x1335, 0x1B35, '\u00A5')  # ¥ yen
  CURRENCY_SIGN = (0x1336, 0x1B36, '\u00A4')  # ¤ non-specific currency sign
  BOX_DRAWINGS_HEAVY_VERTICAL = (0x1337, 0x1B37, '\u2503')  # | Vertical bar

  # Danish extended characters
  LATIN_CAPITAL_LETTER_A_WITH_RING_ABOVE = (0x1338, 0x1B38, '\u00C5')  # Å capital A with ring
  LATIN_SMALL_LETTER_A_WITH_RING_ABOVE = (0x1339, 0x1B39, '\u00E5')  # å small a with ring
  LATIN_CAPITAL_LETTER_O_WITH_STROKE = (0x133A, 0x1B3A, '\u00D8')  # Ø capital O with slash
  LATIN_SMALL_LETTER_O_WITH_STROKE = (0x133B, 0x1B3B, '\u00F8')  # ø small o with slash
  BOX_DRAWINGS_HEAVY_DOWN_AND_RIGHT = (0x133C, 0x1B3C, '\u250F')  # ⎡ upper left corner
  BOX_DRAWINGS_HEAVY_DOWN_AND_LEFT = (0x133D, 0x1B3D, '\u2513')  # ⎤ upper right corner
  BOX_DRAWINGS_HEAVY_UP_AND_RIGHT = (0x133E, 0x1B3E, '\u2517')  # ⎣ lower left corner
  BOX_DRAWINGS_HEAVY_UP_AND_LEFT = (0x133F, 0x1B3F, '\u251B')  # ⎦ lower right corner

  def __init__(self, channel_1: int, channel_2: int, unicode: chr):
    super().__init__(channel_1, channel_2)
    self._unicode = unicode

  def get_unicode_value(self) -> chr:
    """Returns the special or extended character unicode value"""
    return self._unicode

  @staticmethod
  def find(value: int) -> typing.Optional[SccExtendedCharacter]:
    """Find the special character corresponding to the specified value"""
    for spec_char in list(SccExtendedCharacter):
      if spec_char.contains_value(value):
        return spec_char
    return None



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for ISD generation using caching'''

# pylint: disable=R0201,C0115,C0116

from fractions import Fraction
import unittest
from ttconv.isd import ISD
import ttconv.model as model
import ttconv.style_properties as styles
import xml.etree.ElementTree as et
import ttconv.imsc.reader as imsc_reader


class ISDCacheTests(unittest.TestCase):

  def test_large_number_of_regions(self):
    doc = model.ContentDocument()
    body = model.Body(doc)
    doc.set_body(body)

    for i in range(1000):
      r_id = "r" + str(i)
      r = model.Region(r_id, doc)
      doc.put_region(r)

      div = model.Div(doc)
      div.set_region(r)
      body.push_child(div)

      p = model.P(doc)
      p.set_begin(Fraction(2*i))
      p.set_end(Fraction(2*i + 1))
      div.push_child(p)

      span = model.Span(doc)
      span.push_child(model.Text(doc, f"div {i} content"))
      p.push_child(span)

    sig_times = ISD.significant_times(doc)

    for t in sig_times:
      isd = ISD.from_model(doc, t, sig_times)
      regions = list(isd.iter_regions())
      if t % 2 == 0:
        self.assertEqual(len(regions), 1)
        region = list(regions)[0]
        body = list(region)[0]
        div = list(body)[0]
        self.assertEqual(len(div), 1)
        self.assertIsInstance(list(div)[0], model.P)

  def test_regions_with_many_p(self):
    doc = model.ContentDocument()
    body = model.Body(doc)
    doc.set_body(body)

    for i in range(100):
      r_id = "r" + str(i)
      r = model.Region(r_id, doc)
      doc.put_region(r)

      div = model.Div(doc)
      div.set_region(r)
      body.push_child(div)

      for j in range(20):
        p = model.P(doc)
        p.set_begin(Fraction(20 * i + j))
        p.set_end(Fraction(20 * i + j + 1))
        div.push_child(p)

        span = model.Span(doc)
        span.push_child(model.Text(doc, f"div {i} p {j} content"))
        p.push_child(span)

    sig_times = ISD.significant_times(doc)

    for t in sig_times:
      isd = ISD.from_model(doc, t, sig_times)
      regions = list(isd.iter_regions())
      self.assertTrue(len(regions) in (0, 1))


  def test_show_background(self):
    ttml_doc = """<tt xml:lang="en"
    xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling"
    xmlns:ttp="http://www.w3.org/ns/ttml#parameter"
    ttp:timeBase="media">
  <head>
    <styling>
      <style xml:id="s1" tts:backgroundColor="#ff0000" />
    </styling>
    <layout>
      <region xml:id="r1" style="s1" tts:extent="50% 50%" tts:origin="0% 0%"/>
      <region xml:id="r2" style="s1" tts:extent="50% 50%" tts:origin="0% 50%"/>
      <region xml:id="r3" style="s1" tts:extent="50% 50%" tts:origin="50% 50%"/>
      <region xml:id="r4" tts:extent="50% 50%" tts:origin="50% 0%"/>
    </layout>
  </head>
  <body>
    <div>
      <p region="r4" begin="00:00:00" end="00:00:00.2" xml:id="p0">ABCDEF</p>
      <p region="r1" begin="00:00:00.25" end="00:00:04" xml:id="p1"><span style="s1">a</span></p>
      <p region="r2" begin="00:00:00.25" end="00:00:04" xml:id="p2"><span style="s1">g</span></p>
      <p region="r3" begin="00:00:00.25" end="00:00:04" xml:id="p3"><span style="s1">m</span></p>
      <p region="r4" begin="00:00:00.25" end="00:00:04" xml:id="p4"><span style="s1">s</span></p>
    </div>
  </body>
</tt>"""

    doc = imsc_reader.to_model(et.ElementTree(et.fromstring(ttml_doc)))

    sig_times = ISD.significant_times(doc)

    isd = ISD.from_model(doc, 0.2, sig_times)
    # only regions r1, r2, r3 should be visible at 0.2s since r4 has no background
    self.assertEqual(len(list(isd.iter_regions())), 3)

    isd = ISD.from_model(doc, 0.1, sig_times)
    # all regions should be visible at 0.1s since "ABCDEF" is shown in r4
    self.assertEqual(len(list(isd.iter_regions())), 4)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''ttconv tt'''

import json
import logging
import os
import sys
import typing
import xml.etree.ElementTree as et
from argparse import ArgumentParser
from enum import Enum
from pathlib import Path
from ttconv.filters.document_filter import DocumentFilter

import ttconv.imsc.reader as imsc_reader
import ttconv.imsc.writer as imsc_writer
import ttconv.scc.reader as scc_reader
import ttconv.scc.writer as scc_writer
import ttconv.srt.writer as srt_writer
import ttconv.srt.reader as srt_reader
import ttconv.stl.reader as stl_reader
import ttconv.vtt.reader as vtt_reader
from ttconv.vtt.config import VTTWriterConfiguration
import ttconv.vtt.writer as vtt_writer
from ttconv.config import GeneralConfiguration
from ttconv.config import ModuleConfiguration
from ttconv.imsc.config import IMSCWriterConfiguration
from ttconv.isd import ISDConfiguration
from ttconv.scc.config import SccReaderConfiguration, SccWriterConfiguration
from ttconv.stl.config import STLReaderConfiguration
from ttconv.srt.config import SRTReaderConfiguration, SRTWriterConfiguration

LOGGER = logging.getLogger("ttconv")

CONFIGURATIONS = [
  GeneralConfiguration,
  IMSCWriterConfiguration,
  ISDConfiguration,
  SccReaderConfiguration,
  SRTReaderConfiguration,
]


class ProgressConsoleHandler(logging.StreamHandler):
  """
  A handler class which allows the cursor to stay on
  one line for selected messages
  """

  class ProgressType(Enum):
    '''Whether the progress is for reading or writing operations
    '''
    read = 1
    write = 2

  def __init__(self):
    self.is_writing_progress_bar = False
    self.last_progress_msg = ""
    self.display_progress_bar = True
    super().__init__()

  def emit(self, record):

    def progress_str(progress_type: ProgressConsoleHandler.ProgressType, percent_progress: float) -> str:
      '''Formats the progress string.'''

      prefix = "Reading:" if progress_type is ProgressConsoleHandler.ProgressType.read else "Writing:"
      suffix = 'Complete'
      length = 50
      fill = '█'
      filled_length = int(length * percent_progress)
      bar_val = fill * filled_length + '-' * (length - filled_length)

      return f'\r{prefix} |{bar_val}| {100 * percent_progress:3.0f}% {suffix}'

    try:
      msg = self.format(record)

      stream = self.stream

      is_progress_bar_record = hasattr(record, 'progress_bar')
      percent_progress = None

      if not self.display_progress_bar and is_progress_bar_record:
        return

      if is_progress_bar_record:
        percent_progress = getattr(record, 'percent_progress')
        msg = progress_str(
          getattr(record, 'progress_bar'),
          min(1.0, abs(float(percent_progress)))
        )
        self.last_progress_msg = msg

      if self.is_writing_progress_bar and not is_progress_bar_record:
        # erase and over write the progress bar
        stream.write('\r')
        length = len(self.last_progress_msg)
        stream.write(' ' * length)

        # go to beginning of the line and write the new messages
        stream.write('\r')
        stream.write(msg)
        stream.write(self.terminator)

        # write the old progress information
        stream.write(self.last_progress_msg)
      elif is_progress_bar_record:
        stream.write(msg)
      else:
        stream.write(msg)
        stream.write(self.terminator)

      if is_progress_bar_record:
        self.is_writing_progress_bar = True
        if percent_progress is not None and float(percent_progress) >= 1.0:
          sys.stdout.write('\r\n')
          self.is_writing_progress_bar = False

      self.flush()

    except (KeyboardInterrupt, SystemExit):
      raise
    except:  # pylint: disable=bare-except
      self.handleError(record)


def progress_callback_read(percent_progress: float):
  '''Callback handler used by reader and writer.'''
  LOGGER.info(
    "%d%%",
    percent_progress,
    extra={
      'progress_bar': ProgressConsoleHandler.ProgressType.read,
      'percent_progress': percent_progress,
    }
  )


def progress_callback_write(percent_progress: float):
  '''Callback handler used by reader and writer.'''
  LOGGER.info(
    "%d%%",
    percent_progress,
    extra={
      'progress_bar': ProgressConsoleHandler.ProgressType.write,
      'percent_progress': percent_progress,
    }
  )


class FileTypes(Enum):
  '''Enumerates the types of supported'''
  TTML = "ttml"
  SCC = "scc"
  SRT = "srt"
  STL = "stl"
  VTT = "vtt"

  @staticmethod
  def get_file_type(file_type: str, file_extension: str):
    """Convenience function to convert string ased file type 
    and extension to FileTypes."""

    if file_type is None and file_extension is None:
      return None

    if file_type is None:
      if len(file_extension) > 0 and file_extension[0] == '.':
        file_extension = file_extension[1:len(file_extension)]

      return FileTypes(file_extension.lower())

    return FileTypes(file_type.lower())


def read_config_from_json(config_class, json_data) -> typing.Optional[ModuleConfiguration]:
  """Returns a requested configuration from json data"""
  if config_class is None or json_data is None:
    return None

  json_config = json_data.get(config_class.name())

  if json_config is None:
    return None

  return config_class.parse(json_config)


# Argument parsing setup
#
cli = ArgumentParser()
subparsers = cli.add_subparsers(dest="subcommand")


def argument(*name_or_flags, **kwargs):
  """Convenience function to properly format arguments to pass to the
  subcommand decorator."""

  return (list(name_or_flags), kwargs)


def subcommand(args=None, parent=subparsers):
  """Decorator to define a new subcommand in a sanity-preserving way.
  The function will be stored in the ``func`` variable when the parser
  parses arguments so that it can be called directly like so::

      args = cli.parse_args()
      args.func(args)

  Usage example::

      @subcommand([argument("-d", help="Enable debug mode", action="store_true")])
      def subcommand(args):
          print(args)

  Then on the command line::

      $ python cli.py subcommand -d

  """

  def decorator(func):
    parser = parent.add_parser(func.__name__, description=func.__doc__)
    for arg in args:
      parser.add_argument(*arg[0], **arg[1])
    parser.set_defaults(func=func)

  if args is None:
    args = []
  return decorator


@subcommand([
  argument("-i", "--input", help="Input file path", required=True),
  argument("-o", "--output", help="Output file path", required=True),
  argument("--itype", help="Input file type", required=False),
  argument("--otype", help="Output file type", required=False),
  argument("--filter", action="append", help="Document filter", required=False, default=[]),
  argument("--config", help="Configuration in json. Overridden by --config_file.", required=False),
  argument("--config_file", help="Configuration file. Overrides --config.", required=False)
])
def convert(args):
  '''Process input and output through the reader, converter, and writer'''

  inputfile = args.input
  outputfile = args.output

  # Note - Loading config data from a file takes priority over 
  # data passed in as a json string
  json_config_data = None

  if args.config is not None:
    json_config_data = json.loads(args.config)
  if args.config_file is not None:
    with open(args.config_file) as json_file:
      json_config_data = json.load(json_file)

  general_config : GeneralConfiguration = read_config_from_json(GeneralConfiguration, json_config_data)

  if general_config is not None:

    if general_config.progress_bar is not None:
      progress.display_progress_bar = general_config.progress_bar

    if general_config.log_level is not None:
      LOGGER.setLevel(general_config.log_level)

  LOGGER.info("Input file is %s", inputfile)
  LOGGER.info("Output file is %s", outputfile)

  _input_filename, input_file_extension = os.path.splitext(inputfile)
  _output_filename, output_file_extension = os.path.splitext(outputfile)

  reader_type = FileTypes.get_file_type(args.itype, input_file_extension)
  writer_type = FileTypes.get_file_type(args.otype, output_file_extension)

  if reader_type is FileTypes.TTML:
    # 
    # Parse the xml input file into an ElementTree
    #
    tree = et.parse(inputfile)

    #
    # Pass the parsed xml to the reader
    #
    model = imsc_reader.to_model(tree, progress_callback_read)

  elif reader_type is FileTypes.SCC:
    file_as_str = Path(inputfile).read_text()

    #
    # Read the config
    #
    reader_config = read_config_from_json(SccReaderConfiguration, json_config_data)

    #
    # Pass the parsed xml to the reader
    #
    model = scc_reader.to_model(file_as_str, reader_config, progress_callback_read)

  elif reader_type is FileTypes.STL:
    #
    # Read the config
    #
    reader_config = read_config_from_json(STLReaderConfiguration, json_config_data)

    #
    # Open the file and pass it to the reader
    #
    with open(inputfile, "rb") as f:
      model = stl_reader.to_model(f, reader_config, progress_callback_read)

  elif reader_type is FileTypes.SRT:
    #
    # Read the config
    #
    reader_config = read_config_from_json(SRTReaderConfiguration, json_config_data)

    #
    # Open the file and pass it to the reader
    #
    with open(inputfile, "r", encoding="utf-8") as f:
      model = srt_reader.to_model(f, reader_config, progress_callback_read)

  elif reader_type is FileTypes.VTT:

    #
    # Open the file and pass it to the reader
    #
    with open(inputfile, "r", encoding="utf-8") as f:
      model = vtt_reader.to_model(f, None, progress_callback_read)

  else:
    if args.itype is not None:
      exit_str = f'Input type {args.itype} is not supported'
    else:
      exit_str = f'Input file {args.input} is not supported'

    LOGGER.error(exit_str)
    sys.exit(exit_str)

  #
  # apply document language
  #
  if general_config is not None and general_config.document_lang is not None:
    model.set_lang(general_config.document_lang)

  #
  # apply document filter
  #

  for filter_name in args.filter:
    doc_filter_class = DocumentFilter.get_filter_by_name(filter_name)

    if doc_filter_class is None:
      LOGGER.error("Unknown filter: %s", filter_name)
      continue

    filter_config_class = doc_filter_class.get_config_class()

    filter_config = read_config_from_json(filter_config_class, json_config_data)

    doc_filter: DocumentFilter = doc_filter_class(filter_config or filter_config_class())

    doc_filter.process(model)

  #
  # Write the output document
  #
  if writer_type is FileTypes.TTML:
    #
    # Read the config
    #
    writer_config = read_config_from_json(IMSCWriterConfiguration, json_config_data)

    #
    # Construct and configure the writer
    #
    tree_from_model = imsc_writer.from_model(model, writer_config, progress_callback_write)

    #
    # Write out the converted file
    #
    tree_from_model.write(outputfile, encoding="utf-8")

  elif writer_type is FileTypes.SRT:
    #
    # Read the config
    #
    writer_config = read_config_from_json(SRTWriterConfiguration, json_config_data)

    #
    # Construct and configure the writer
    #
    srt_document = srt_writer.from_model(model, writer_config, progress_callback_write)

    #
    # Write out the converted file
    #
    with open(outputfile, "w", encoding="utf-8") as srt_file:
      srt_file.write(srt_document)

  elif writer_type is FileTypes.VTT:
    #
    # Read the config
    #
    writer_config = read_config_from_json(VTTWriterConfiguration, json_config_data)

    #
    # Construct and configure the writer
    #
    vtt_document = vtt_writer.from_model(model, writer_config, progress_callback_write)

    #
    # Write out the converted file
    #
    with open(outputfile, "w", encoding="utf-8") as vtt_file:
      vtt_file.write(vtt_document)

  elif writer_type is FileTypes.SCC:
    #
    # Read the config
    #
    writer_config = read_config_from_json(SccWriterConfiguration, json_config_data)

    #
    # Construct and configure the writer
    #
    scc_document = scc_writer.from_model(model, writer_config, progress_callback_write)

    #
    # Write out the converted file
    #
    with open(outputfile, "w", encoding="utf-8") as scc_file:
      scc_file.write(scc_document)


  else:
    if args.otype is not None:
      exit_str = f'Output type {args.otype} is not supported'
    else:
      exit_str = f'Output file is {args.output} is not supported'

    LOGGER.error(exit_str)
    sys.exit(exit_str)


# Ensure that the handler is added only once/globally
# Otherwise the handler will be called multiple times
progress = ProgressConsoleHandler()
LOGGER.addHandler(progress)
LOGGER.setLevel(logging.INFO)


def main(argv=None):
  '''Main application processing'''

  args = cli.parse_args(argv if argv is not None else sys.argv[1:])
  if args.subcommand is None:
    cli.print_help()
  else:
    args.func(args)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the IMSC \\<fontFamily\\> parser'''

# pylint: disable=R0201,C0115,C0116

import unittest
from ttconv.imsc.utils import parse_font_families, serialize_font_family
import ttconv.style_properties as styles

class IMSCReaderTest(unittest.TestCase):

  _parse_tests = [
    ["default", [styles.GenericFontFamilyType.default]],
    ["'default'", ["default"]],
    ["foo, 'bar good'", ["foo", "bar good"]],
    ['foo, "bar good"', ["foo", "bar good"]],
    [r'foo, "bar \good"', ["foo", "bar good"]],
    [r'foo, "bar \,good"', ["foo", "bar ,good"]],
    [r'foo, "bar,good"', ["foo", "bar,good"]]
  ]

  def test_parse_font_families(self):
    for test in self._parse_tests:
      with self.subTest(test[0]):
        c = parse_font_families(test[0])
        self.assertEqual(c, test[1])


  _serialize_tests = [
    [(styles.GenericFontFamilyType.default,), "default"],
    [("default",), '"default"'],
    [(styles.GenericFontFamilyType.proportionalSansSerif, "bar good"), 'proportionalSansSerif, "bar good"'],
    [("foo", "bar, good"), r'"foo", "bar, good"'],
    [("bar\"good",), r'"bar\"good"']
  ]

  def test_serialize_font_families(self):
    for test in self._serialize_tests:
      with self.subTest(test[0]):
        c = serialize_font_family(test[0])
        self.assertEqual(c, test[1])

if __name__ == '__main__':
  unittest.main()
